#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on October 26, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from functools import partial

from qtpy.QtCore import QItemSelection, QModelIndex, Qt, Slot
from qtpy.QtGui import QClipboard, QCursor, QKeySequence
from qtpy.QtWidgets import (
    QAction, QApplication, QDialog, QHeaderView, QMessageBox, QTreeView)

import karabogui.icons as icons
from karabo.common.project.api import find_parent_object
from karabogui import messagebox
from karabogui.access import AccessRole, access_role_allowed
from karabogui.events import KaraboEvent, broadcast_event
from karabogui.project.dialog.project_handle import NewProjectDialog
from karabogui.project.utils import maybe_save_modified_project, save_object
from karabogui.singletons.api import (
    get_db_conn, get_project_model, get_selection_tracker)
from karabogui.util import is_database_processing, move_to_cursor

from .controller.bases import BaseProjectGroupController
from .controller.device import DeviceInstanceController
from .controller.macro import MacroController, get_project_macros
from .controller.project import ProjectController
from .controller.project_groups import ProjectSubgroupController
from .controller.server import DeviceServerController, get_project_servers
from .controller.subproject import SubprojectController

EXPAND_DEPTH = 2


class ProjectView(QTreeView):
    """ An object representing the view for a Karabo project
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setObjectName("projectTreeView")
        project_model = get_project_model()
        project_model.setParent(self)

        self.setModel(project_model)
        project_model.rowsInserted.connect(self._items_added)
        self.setSelectionModel(project_model.q_selection_model)
        self.selectionModel().selectionChanged.connect(self._selection_change)
        self.setDragEnabled(True)

        header = self.header()
        header.moveSection(1, 0)
        header.setSectionResizeMode(0, QHeaderView.Stretch)
        header.setSectionResizeMode(1, QHeaderView.Fixed)
        header.setStretchLastSection(False)
        icon_size = 26
        header.setMaximumSectionSize(icon_size)
        header.resizeSection(1, icon_size)

        # Prevent drag reorder of the header
        header.setSectionsMovable(False)

        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self._show_context_menu)
        self.expanded = False
        self.header().sectionDoubleClicked.connect(self.onDoubleClickHeader)

        self.setUniformRowHeights(True)

    # ----------------------------
    # Public methods

    def destroy_model(self):
        """ Unset project's data model before death.
        """
        model = self.model()
        model.root_model = None

    # ----------------------------
    # Qt methods

    def closeEvent(self, event):
        model = self.model()
        model.root_model = None
        model.setParent(None)
        event.accept()

    def keyPressEvent(self, event):
        if event.matches(QKeySequence.Copy):
            controller = self._get_selected_controller()
            if controller is not None:
                clipboard = QApplication.clipboard()
                # Erase selection clipboard first!
                clipboard.clear(mode=QClipboard.Selection)
                clipboard.setText(controller.model.uuid,
                                  mode=QClipboard.Selection)
                event.accept()
                return
        elif event.matches(QKeySequence.Delete):
            selected_controller = self._get_selected_controller()
            if selected_controller is not None:
                project_controller = self._project_controller(
                    selected_controller)
                selected_controller.delete_press(project_controller,
                                                 parent=self)
                event.accept()
                return

        super().keyPressEvent(event)

    def mouseDoubleClickEvent(self, event):
        selected_controller = self._get_selected_controller()
        if selected_controller is not None:
            project_controller = self._project_controller(selected_controller)
            selected_controller.double_click(project_controller, parent=self)

            ignored = (DeviceInstanceController, DeviceServerController,
                       MacroController)
            if isinstance(selected_controller, ignored):
                # Don't expand or fold servers and devices!
                event.ignore()
                return

        super().mouseDoubleClickEvent(event)

    # ----------------------------
    # Slots

    @Slot(QModelIndex, int, int)
    def _items_added(self, parent_index, start, end):
        """React to the addition of an item (or items).
        """
        # Bail immediately if not the first item
        if start != 0:
            return

        controller = parent_index.internalPointer()
        if (controller is not None and
                isinstance(controller, BaseProjectGroupController) and not
                isinstance(controller, DeviceInstanceController)):
            # If a group just added its first item, expand it
            # except for device instances
            self.expand(parent_index)

    @Slot(QItemSelection, QItemSelection)
    def _selection_change(self, selected, deselected):
        """ Notify controller objects when their Qt list item object is
        selected.
        """
        selected_controller = self._get_selected_controller()
        if selected_controller is not None:
            project_controller = self._project_controller(selected_controller)
            selected_controller.single_click(project_controller, parent=self)

            # Grab control of the global selection
            get_selection_tracker().grab_selection(self.selectionModel())

    @Slot()
    def _show_context_menu(self):
        """ Show a context menu for the currently selected item.
        """
        # Make sure there are not pending DB things in the pipe
        if is_database_processing():
            return

        selected_controller = self._get_selected_controller()
        if selected_controller is not None:
            project_allowed = access_role_allowed(AccessRole.PROJECT_EDIT)
            service_allowed = access_role_allowed(AccessRole.SERVICE_EDIT)

            project_controller = self._project_controller(selected_controller)
            menu = selected_controller.context_menu(project_controller,
                                                    parent=self)
            if isinstance(selected_controller, ProjectController):
                selected_project = selected_controller.model
                rename_action = QAction(icons.edit, 'Rename', menu)
                rename_action.triggered.connect(partial(self._rename_project,
                                                        selected_project))
                rename_action.setEnabled(project_allowed)

                save_action = QAction(icons.save, 'Save', menu)
                save_action.triggered.connect(partial(save_object,
                                                      selected_project,
                                                      domain=None))
                save_action.setEnabled(project_allowed)

                close_action = QAction(icons.kill, 'Close project', menu)
                close_action.triggered.connect(partial(self._close_project,
                                                       selected_project,
                                                       project_controller,
                                                       show_dialog=True))

                instantiate_all_devices = QAction(
                    icons.run, 'Instantiate all devices', menu)
                instantiate_all_devices.triggered.connect(
                    partial(self._instantiate_devices, selected_controller))
                instantiate_all_devices.setEnabled(service_allowed)

                instantiate_all_macros = QAction(
                    icons.run, 'Instantiate all macros', menu)
                instantiate_all_macros.triggered.connect(
                    partial(self._instantiate_macros, selected_controller))
                instantiate_all_macros.setEnabled(service_allowed)

                is_trashed = selected_project.is_trashed
                if is_trashed:
                    text = 'Restore from trash'
                else:
                    text = 'Move to trash'
                trash_action = QAction(icons.delete, text, menu)
                trash_action.triggered.connect(partial(self.update_is_trashed,
                                                       selected_project,
                                                       project_controller))
                trash_action.setEnabled(project_allowed)
                menu.addAction(rename_action)
                menu.addAction(instantiate_all_macros)
                menu.addAction(instantiate_all_devices)
                menu.addSeparator()
                menu.addAction(save_action)
                menu.addAction(close_action)
                menu.addSeparator()
                menu.addAction(trash_action)

            menu.exec(QCursor.pos())

    # ----------------------------
    # Private methods

    def _get_selected_controller(self):
        """ Return the currently selected controller.
        """
        indices = self.selectionModel().selectedIndexes()
        if not indices:
            return None

        first_index = indices[0]
        return first_index.internalPointer()

    def _project_controller(self, controller):
        """ Find the parent project controller of a given controller
        """
        parent_types = (ProjectController, ProjectSubgroupController,
                        SubprojectController)
        root_controller = self.model().root_controller
        return find_parent_object(controller, root_controller, parent_types)

    def _rename_project(self, project):
        """ Change the ``simple_name`` of the given ``project``
        """
        dialog = NewProjectDialog(model=project, is_rename=True, parent=self)
        move_to_cursor(dialog)
        result = dialog.exec()
        if result == QDialog.Accepted:
            project.simple_name = dialog.simple_name

    @Slot()
    def _instantiate_devices(self, selected_controller):
        """ Instantiate all devices in the given project

        The project controller ``selected_controller`` is used to find the
        children server controller to instantiate devices
        """
        online, offline = get_project_servers(
            selected_controller)
        # Check for offline servers to provide a warning
        for server in online:
            server.instantiate_devices()

        # Notify that we have seen offline servers!
        if offline:
            offline_ids = ', '.join(
                sorted({server.model.server_id
                        for server in offline}))
            msg = ('Servers are not online for device instantiation: '
                   '{}!'.format(offline_ids))
            messagebox.show_warning(msg, title='Servers offline',
                                    parent=self)

    @Slot()
    def _instantiate_macros(self, selected_controller):
        """ Instantiate all macros in the given project

        The project controller ``selected_controller`` is used to find the
        offline macro controller to instantiate macros
        """
        _, offline = get_project_macros(selected_controller)
        for macro in offline:
            macro.run_macro()

    def _close_project(self, project, project_controller, show_dialog=False):
        """ Close the given `project`
        """
        if show_dialog:
            ask = ('Do you really want to close the project \"<b>{}</b>\"'
                   '?').format(project.simple_name)
            reply = QMessageBox.question(self, 'Close project', ask,
                                         QMessageBox.Yes | QMessageBox.No,
                                         QMessageBox.No)
            if reply == QMessageBox.No:
                return

        if project_controller is not None:
            # Check for modififications before closing
            if show_dialog and not maybe_save_modified_project(project,
                                                               parent=self):
                return
            # A subproject
            parent_project_model = project_controller.model
            if project in parent_project_model.subprojects:
                parent_project_model.subprojects.remove(project)
        else:
            # Check for modififications before closing
            model = self.model().root_model
            if show_dialog and not maybe_save_modified_project(model,
                                                               parent=self):
                return
            # The master project
            self.model().root_model = None
            broadcast_event(KaraboEvent.ProjectFilterUpdated,
                            {'status': False})
            # And reset the object name
            broadcast_event(KaraboEvent.ProjectName,
                            {"simple_name": None})

    def update_is_trashed(self, project, project_controller):
        """ Mark the given `project` as (un-)trashed
        """
        action = "untrash" if project.is_trashed else "trash"
        ask = ('Are you sure you want to <b>{}</b> the project'
               ' \"<b>{}</b>\"?'.format(action, project.simple_name))
        msg_box = QMessageBox(QMessageBox.Question, 'Change of project state',
                              ask, QMessageBox.Yes | QMessageBox.No,
                              parent=self)
        msg_box.setModal(False)
        msg_box.setDefaultButton(QMessageBox.No)
        if msg_box.exec() == QMessageBox.Yes:
            project.is_trashed = not project.is_trashed
            db_conn = get_db_conn()
            db_conn.update_trashed(db_conn.default_domain, 'project',
                                   project.uuid, project.is_trashed)
            # We directly save on attribute update!
            save_object(project)

    @Slot()
    def onDoubleClickHeader(self):
        if self.expanded:
            self.collapseAll()
        else:
            self.expandAll()

    @Slot()
    def resetExpand(self):
        self.expandAll()

    def collapseAll(self):
        self.expanded = False
        super().collapseAll()

    def expandAll(self):
        self.expanded = True
        super().expandToDepth(EXPAND_DEPTH)



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on October 27, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtWidgets import QMenu
from traits.api import Instance, List, on_trait_change

from karabo.common.project.api import ProjectModel
from karabogui import icons, messagebox
from karabogui.fonts import get_qfont
from karabogui.itemtypes import ProjectItemTypes

from .bases import BaseProjectController
from .project_groups import ProjectControllerUiData, ProjectSubgroupController


class ProjectController(BaseProjectController):
    """ A controller for ProjectModel objects
    """
    # Redefine model with the correct type
    model = Instance(ProjectModel)
    # The subgroups of this project
    children = List(Instance(ProjectSubgroupController))

    def context_menu(self, project_controller, parent=None):
        return QMenu(parent)

    def create_ui_data(self):
        font = get_qfont()
        font.setBold(True)
        return ProjectControllerUiData(font=font, icon=icons.folder)

    @on_trait_change('model.is_trashed,model.conflict')
    def _update_icon_style(self):
        if self.model.is_trashed:
            self.ui_data.icon = icons.folderTrash
            if self.parent is not None:
                messagebox.show_warning(
                    "You are working with an actively trashed project "
                    "<b>{}</b>! Please untrash or close "
                    "the project.".format(self.model.simple_name))
        elif self.model.conflict:
            self.ui_data.icon = icons.alarmGlobal
        else:
            self.ui_data.icon = icons.folder

    def child(self, index):
        """Returns a child of this controller.

        :param index: An index into the list of this controller's children
        :returns: A BaseProjectController instance or None
        """
        return self.children[index]

    def rows(self):
        """Returns the number of rows 'under' this controller in the project
        tree view.
        """
        return len(self.children)

    def info(self):
        return {'type': ProjectItemTypes.PROJECT}



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from karabo.native import Hash
from karabogui.testing import GuiTestCase
from karabogui.widgets.popup import PopupWidget


class TestPopUp(GuiTestCase):

    def test_default_widget(self):
        widget = PopupWidget()
        # Freeze is `False` as default
        layout = widget.layout()
        assert not layout.isEmpty()
        assert getattr(widget, '_ui_freeze_button', None) is None
        assert widget.freeze is False

        # Test the info setting
        info = {"minInc": 5, "description": "This is a description"}
        widget.setInfo(info)
        plain_text = "\nminInc: \n5\ndescription: \nThis is a description\n"
        assert widget.text == plain_text

        info = Hash("minInc", 5, "description", "This is a description")
        widget.setInfo(info)
        # Hashes are different
        plain_text = "\nminInc\n5\ndescription\nThis is a description\n"
        assert widget.text == plain_text
        widget.close()

    def test_freeze_widget(self):
        widget = PopupWidget(can_freeze=True)
        layout = widget.layout()
        assert not layout.isEmpty()
        assert getattr(widget, '_ui_freeze_button', None) is not None
        assert widget.freeze is False

        # Test the info setting
        info = {"minInc": 20, "description": "This is freeze"}
        widget.setInfo(info)
        plain_text = "\nminInc: \n20\ndescription: \nThis is freeze\n"
        assert widget.text == plain_text

        # Toggle freeze and test the caching
        widget.toggle_freeze()
        assert widget.freeze is True
        # Test the info setting
        info = {"alarmLow": 20, "displayedName": "Alarm Info"}
        widget.setInfo(info)
        # Text did not change... But if we toggle, we change again!
        assert widget.text == plain_text
        widget.toggle_freeze()
        assert widget.freeze is False
        plain_text = "\nalarmLow: \n20\ndisplayedName: \nAlarm Info\n"
        assert widget.text == plain_text

        # Test the reset
        widget.toggle_freeze()
        assert widget.freeze is True
        widget.reset()
        assert widget.freeze is False
        assert widget.text == plain_text
        widget.close()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtCore import Qt

from karabogui.dialogs.api import ApplicationConfigurationDialog
from karabogui.singletons.configuration import Configuration
from karabogui.testing import GuiTestCase, singletons


class TestAppConfDialog(GuiTestCase):

    def test_basic_dialog(self):
        config = Configuration()
        with singletons(configuration=config):
            dialog = ApplicationConfigurationDialog()
            assert not dialog.isModal()
            model = dialog.tree_view.model()
            assert model is not None
            # We have 6 childen on root (groups)
            assert model.rowCount() == 5
            group_index = model.index(1, 0)
            index = model.index(0, 0, group_index)
            assert index.data() is not None
            index = model.index(0, 1, group_index)
            assert index.data() is not None

            assert model.headerData(
                0, Qt.Horizontal, Qt.DisplayRole) == 'Name'
            assert model.headerData(
                1, Qt.Horizontal, Qt.DisplayRole) == 'Setting'

            flag = model.flags(index)
            assert int(flag) == 33
            assert flag & Qt.ItemIsEnabled == Qt.ItemIsEnabled
            assert flag & Qt.ItemIsSelectable == Qt.ItemIsSelectable
            assert flag & Qt.ItemIsEditable != Qt.ItemIsEditable
            assert model.columnCount(None) == 2

            assert model.setData(index, "Karabo", Qt.EditRole)
            assert index.data() == "Karabo"

            # Check a boolean
            group_index = model.index(4, 0)
            assert group_index.data() == "user"
            index = model.index(9, 0, group_index)
            assert index.data() == "wizard"
            index_value = model.index(9, 1, group_index)
            flag = model.flags(index_value)
            assert int(flag) == 49
            assert flag & Qt.ItemIsEnabled == Qt.ItemIsEnabled
            assert flag & Qt.ItemIsUserCheckable == Qt.ItemIsUserCheckable

            assert not dialog.expanded
            dialog.expandAll()
            assert dialog.expanded

            # Double clicks, unfortunately, we cannot use QTest for that
            dialog.onDoubleClickHeader()
            assert not dialog.expanded
            dialog.onDoubleClickHeader()
            assert dialog.expanded

    def test_model_tester(self):
        from pytestqt.modeltest import ModelTester
        config = Configuration()
        with singletons(configuration=config):
            dialog = ApplicationConfigurationDialog()
            model = dialog.tree_view.model()
            tester = ModelTester(None)
            tester.check(model)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from collections import namedtuple

from karabogui.programs.theatre import DEVSCENE_PROG, create_theatre
from karabogui.singletons.network import Network
from karabogui.testing import (
    set_test_organization_info, singletons, system_hash)
from karabogui.topology.api import SystemTopology
from karabogui.util import process_qt_events


def test_device_link_regex():
    match = DEVSCENE_PROG.match('SCOPE_GROUP_COMPONENT-2/TYPE/MEMBER|scene')
    assert (match and all(match.groups()))
    match = DEVSCENE_PROG.match('random-string')
    assert not (match and all(match.groups()))


namespace = namedtuple("namespace", "host port nosplash "
                                    "scene_id timeout")


def test_theatre_normal(gui_app, mocker):
    """Test the karabo theatre with a topology"""
    ns = namespace("myhost", "myport", True, ["divvy|scene"], 1)

    network = Network()
    network.connectToServerDirectly = mocker.Mock()
    network.connectToServerDirectly.return_value = True
    network.onSubscribeLogs = mocker.Mock()
    info = mocker.patch("karabogui.programs.base.set_app_info")

    topology = SystemTopology()
    path = "karabogui.programs.theatre.get_scene_from_server"
    with singletons(network=network, topology=topology):
        scene = mocker.patch(path)
        success, waiter, app = create_theatre(ns)
        info.assert_called_with(app, "XFEL", "xfel.eu", "KaraboGUI")
        # Make sure no settings are erased in tests
        set_test_organization_info(app)

        assert success
        assert len(waiter.device_scenes) == 1
        network.signalServerConnectionChanged.emit(True)

        network.onSubscribeLogs.assert_not_called()
        # Initialize the trees
        topology.initialize(system_hash())
        process_qt_events(timeout=1000)
        scene.assert_called_with("divvy", "scene")
    network.deleteLater()


def test_theatre_timeout(gui_app, mocker):
    """Test the karabo theatre with a timeout for the topology"""
    ns = namespace("myhost", "myport", True, ["divvy|scene"], 0)

    network = Network()
    network.connectToServerDirectly = mocker.Mock()
    network.connectToServerDirectly.return_value = True
    network.onSubscribeLogs = mocker.Mock()

    topology = SystemTopology()
    scene_path = "karabogui.programs.theatre.get_scene_from_server"
    box_path = "karabogui.programs.theatre.messagebox"
    with singletons(network=network, topology=topology):
        scene = mocker.patch(scene_path)
        mbox = mocker.patch(box_path)
        success, waiter, app = create_theatre(ns)
        # Make sure no settings are erased in tests
        set_test_organization_info(app)

        assert success
        assert len(waiter.device_scenes) == 1
        network.signalServerConnectionChanged.emit(True)

        process_qt_events(timeout=1000)
        scene.assert_not_called()
        mbox.show_warning.assert_called_once()
    network.deleteLater()


def test_missing_device(gui_app, mocker):
    """Verify the warning message is displyed when device is invalid or has
    no scene."""
    ns = namespace("myhost", "myport", True, ["divvy|scene"], 1)

    network = Network()
    network.connectToServerDirectly = mocker.Mock()
    network.connectToServerDirectly.return_value = True
    network.onSubscribeLogs = mocker.Mock()

    topology = SystemTopology()
    waring_dialog = "karabogui.programs.theatre.messagebox.show_warning"
    with singletons(network=network, topology=topology):
        success, waiter, app = create_theatre(ns)
        set_test_organization_info(app)

        mbox = mocker.patch(waring_dialog)

        # Device doesn't exist in the Topology.
        waiter._topo_update()
        msg = ("The following deviceIds are not present in the system "
               "topology: 'divvy'. \n\nPlease review command line arguments "
               "accordingly.")
        assert mbox.call_count == 1
        mbox.assert_called_with(msg, title="Theater")

        mbox.reset_mock()
        waiter.no_scenes = False
        waiter._topo_update()
        assert mbox.call_count == 0

        # Device with no Scene
        waiter.not_capable_devices = ["FooClass"]
        waiter._topo_update()
        msg = ("The following deviceIds do not provide scenes: 'FooClass'. "
               "\n\nPlease review command line arguments accordingly.")
        assert mbox.call_count == 1
        mbox.assert_called_with(msg, title="Theater")



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from karabogui.panels.navigationpanel import TopologyPanel
from karabogui.singletons.api import get_config
from karabogui.singletons.configuration import Configuration
from karabogui.testing import singletons, system_hash
from karabogui.topology.api import SystemTopology


def test_topology_panel(gui_app, mocker):
    mediator = mocker.Mock()
    topology = SystemTopology()
    topology.initialize(system_hash())
    config = Configuration()
    with singletons(mediator=mediator, topology=topology,
                    configuration=config):
        panel = TopologyPanel()
        assert repr(panel) == "<TopologyPanel topic=None>"
        get_config()["broker_topic"] = "Test"
        assert repr(panel) == "<TopologyPanel topic=Test>"
        assert panel.tree_view.model().rowCount() > 0


def test_service_manager_button(gui_app, mocker):
    panel = TopologyPanel()
    panel._event_network({'status': True})
    assert panel.daemon_button.isVisibleTo(panel)

    panel._event_network({'status': False})
    assert not panel.daemon_button.isVisibleTo(panel)

    panel._event_network({'status': True})
    assert panel.daemon_button.isVisibleTo(panel)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from traits.api import pop_exception_handler, push_exception_handler

from karabo.common.api import InstanceStatus
from karabogui.itemtypes import NavigationItemTypes

from ..device_tree import DeviceTreeNode


def setUp():
    push_exception_handler(lambda *args: None, reraise_exceptions=True)


def tearDown():
    pop_exception_handler()


def test_tree_node_basics():
    empty = DeviceTreeNode(
        node_id='node_id',
        status=InstanceStatus.OK,
        children=[]
    )

    assert empty.child('a_child') is None
    assert empty.info() is None
    assert empty.row() == 0


def test_tree_node_levels():
    root = DeviceTreeNode()
    level_0 = DeviceTreeNode(node_id='FXE_OGT1_BIU', parent=root, level=0)
    root.children.append(level_0)
    level_1 = DeviceTreeNode(node_id='MOTOR', parent=level_0, level=1)
    level_0.children.append(level_1)
    level_2 = DeviceTreeNode(node_id='FXE_OGT1_BIU/MOTOR/SCREEN',
                             parent=level_1, level=2)
    level_1.children.append(level_2)

    assert root.child('FXE_OGT1_BIU') is level_0
    assert level_0.child('MOTOR') is level_1 and level_0.row() == 0
    assert (level_1.child('FXE_OGT1_BIU/MOTOR/SCREEN') is level_2
            and level_1.row() == 0)

    level_0_info = level_0.info()
    assert level_0_info['type'] == NavigationItemTypes.UNDEFINED
    assert level_0_info['domainId'] == 'FXE_OGT1_BIU'
    level_1_info = level_1.info()
    assert level_1_info['type'] == NavigationItemTypes.UNDEFINED
    assert level_1_info['typeId'] == 'MOTOR'
    level_2_info = level_2.info()
    assert level_2_info['type'] == NavigationItemTypes.DEVICE
    assert level_2_info['deviceId'] == 'FXE_OGT1_BIU/MOTOR/SCREEN'



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from pyqtgraph import PlotDataItem
from qtpy.QtCore import QPointF, Qt
from qtpy.QtGui import QBrush, QColor, QLinearGradient, QPen
from traits.api import (
    Array, Constant, Instance, Property, String, cached_property,
    on_trait_change)

from karabogui.graph.common.api import COLORMAPS
from karabogui.graph.image.utils import rescale

from ..base.plot import BasePlot


class HistogramPlot(BasePlot):

    # The x- and y-labels of the plot
    x_label = Constant("Intensity")
    y_label = Constant("Counts")

    # Image levels
    levels = Array(value=[0, 255])

    # Colormap of the
    colormap = String("viridis")

    # --- plot items ---
    # The line item of the plot
    _data_item = Instance(PlotDataItem)

    # --- UI attributes ---
    # The default pen of the plot.
    _pen = Instance(QPen, args=(Qt.NoPen,))

    # The gradient of the brush to fill the area under the curve
    gradient = Property(Instance(QLinearGradient), depends_on='colormap')

    def __init__(self, **traits):
        super().__init__(**traits)
        self._data_item = self._add_plot_item()

    # -----------------------------------------------------------------------
    # Public methods

    def set_data(self, x_data, y_data):
        # Retrieve the gradient range from the input x_data and image levels
        brush = self._calc_brush(x_data)

        # plot the data with the resulting gradient
        self._data_item.setData(x_data, y_data, stepMode="center",
                                pen=self._pen, brush=brush, fillLevel=0)

    def destroy(self):
        super().destroy()
        self.on_trait_change(self._set_brush, "levels", remove=True)

    # -----------------------------------------------------------------------
    # Private methods

    def _calc_brush(self, x_data):
        start, stop = rescale(self.levels,
                              min_value=x_data[0],
                              max_value=x_data[-1],
                              low=0, high=1)

        # Create the brush by modifying the gradient ranges
        gradient = self.gradient
        gradient.setStart(round(start, 3), 0)
        gradient.setFinalStop(round(stop, 3), 0)
        return QBrush(gradient)

    # -----------------------------------------------------------------------
    # Trait handlers

    @on_trait_change("levels")
    def _set_brush(self):
        # Get brush with the new levels and from existing data
        x_data = self._data_item.xData
        if x_data is not None and x_data.size:
            brush = self._calc_brush(x_data)
            self._data_item.setBrush(brush)

    # -----------------------------------------------------------------------
    # Trait properties

    @cached_property
    def _get_gradient(self):
        gradient = QLinearGradient(QPointF(0, 0), QPointF(1, 0))
        gradient.setCoordinateMode(QLinearGradient.ObjectBoundingMode)
        grad_stops = [(stop, QColor(*color))
                      for stop, color in COLORMAPS[self.colormap]]
        gradient.setStops(grad_stops)
        return gradient



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on June 9, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import os.path as op

from qtpy.QtWidgets import QBoxLayout

import karabo.common.scenemodel.tests as sm
from karabo.common.scenemodel.api import BoxLayoutModel, LabelModel, LineModel
from karabogui.sceneview.layout.api import BoxLayout
from karabogui.sceneview.shapes import LineShape
from karabogui.sceneview.widget.api import LabelWidget

DATA_DIR = op.join(op.abspath(op.dirname(sm.__file__)), 'data')


def test_horizontal_box_layouts(gui_app):
    # Horizonal layout
    model = BoxLayoutModel(direction=QBoxLayout.LeftToRight)
    boxLayout = BoxLayout(model, model.direction)
    assert boxLayout.count() == 0
    # Add a child to layout model
    label_model = LabelModel(x=0, y=0, text="foo")
    label_widget = LabelWidget(label_model)
    boxLayout._add_widget(label_widget)
    assert boxLayout.count() == 1
    assert boxLayout.itemAt(0).widget() is label_widget


def test_vertical_box_layout(gui_app):
    # Vertical layout
    model = BoxLayoutModel(direction=QBoxLayout.TopToBottom)
    boxLayout = BoxLayout(model, model.direction)
    assert boxLayout.count() == 0

    model = LineModel(x1=0, y1=0, x2=1, y2=1)
    # Add shape to layout
    lineShape = LineShape(model=model)
    boxLayout._add_shape(lineShape)
    assert boxLayout.count() == 1
    assert boxLayout.itemAt(0).shape is lineShape



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtWidgets import QWizard, QWizardPage
from traits.api import (
    Constant, DelegatesTo, HasStrictTraits, Instance, on_trait_change)

from .pages import ConfigurePage, LinkPage, SelectScenesPage


class BasePageController(HasStrictTraits):
    # Every page has a reference to the main trait wizard
    wizard_controller = Instance(object)

    page = Instance(QWizardPage)
    needs_completion = Constant(False)
    commit_page = Constant(False)

    # Delegates
    available_scenes = DelegatesTo("wizard_controller")
    selected_scenes = DelegatesTo("wizard_controller")
    link = DelegatesTo("wizard_controller")
    project_model = DelegatesTo("wizard_controller")

    def __init__(self, **traits):
        super().__init__(**traits)
        page = self.page
        page.set_complete(not self.needs_completion)
        page.setCommitPage(self.commit_page)

    def init(self):
        """Re-implement on subclass if further processes are needed on
         initialization of page. Visiting it back doesn't trigger this anymore.
        """

    def enter(self):
        """Re-implement on subclass if actions are needed to be done when
        the page is loaded. Visiting it back triggers this."""

    def exit(self):
        """Re-implement on subclass if actions are needed to be done when
        the page is left"""


class SelectScenesController(BasePageController):
    """The scenes controller is for the validation of the selected scenes"""
    page = Instance(SelectScenesPage)
    needs_completion = Constant(True)

    # ----------------------------------------------------------------------
    # Controller methods

    def init(self):
        """Populate the list pages"""
        available = [(scene.simple_name, scene)
                     for scene in self.available_scenes]
        self.page.set_available_scenes(available)

        selected = [(scene.simple_name, scene)
                    for scene in self.selected_scenes]
        self.page.set_selected_scenes(selected)
        self._validate_selected_scenes()

    # ----------------------------------------------------------------------
    # Trait methods

    def _page_default(self):
        page = SelectScenesPage()
        # Note: For some reason on races we cannot use Qt signals here
        # Hence, use handlers.
        page.sceneAdded = self._add_scene
        page.sceneRemoved = self._remove_scene
        return page

    @on_trait_change("selected_scenes_items")
    def _validate_selected_scenes(self):
        has_selected = bool(len(self.selected_scenes))
        self.page.set_complete(has_selected)

    # ----------------------------------------------------------------------
    # Private methods

    def _add_scene(self, model):
        if model is None:
            return

        self.available_scenes.remove(model)
        self.selected_scenes.append(model)

    def _remove_scene(self, model):
        if model is None:
            return

        self.selected_scenes.remove(model)
        self.available_scenes.append(model)


class ConfigureController(BasePageController):
    """The Configure Controller provides the basic configuration options
    for the network, username, splash and login dialog"""
    page = Instance(ConfigurePage)
    needs_completion = Constant(True)
    commit_page = Constant(True)

    def init(self):
        if self.project_model is not None:
            self.page.set_subheader(
                f"Project: {self.project_model.simple_name}")

        self.link["uuids"] = [scene.uuid for scene in self.selected_scenes]
        self.page.set_config(self.link)
        self.page.set_complete(True)

    def enter(self):
        self.page.configChanged.connect(self._set_config)

    def exit(self):
        self.page.configChanged.disconnect(self._set_config)

    # ----------------------------------------------------------------------
    # Trait methods

    def _page_default(self):
        return ConfigurePage()

    # ----------------------------------------------------------------------
    # Helper methods

    def _set_config(self, config):
        self.link.update(config)


class LinkController(BasePageController):
    """The link controller is the final controller setting the LinkPage
    to generate the `url` to the clipboard"""
    page = Instance(LinkPage)

    def init(self):
        self.page.set_data(self.link)

    def enter(self):
        button = self.wizard_controller.widget.button(QWizard.FinishButton)
        button.clicked.connect(self.page.finished)

    def exit(self):
        button = self.wizard_controller.widget.button(QWizard.FinishButton)
        button.clicked.disconnect(self.page.finished)

    # ----------------------------------------------------------------------
    # Trait methods

    def _page_default(self):
        page = LinkPage()
        return page



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import re

from karabo.common.scenemodel.api import HistoricTextModel
from karabo.native import (
    Configurable, Hash, String, Timestamp, UInt32, VectorString)
from karabogui.const import IS_WINDOWS_SYSTEM
from karabogui.controllers.display.historic_text import DisplayHistoricText
from karabogui.testing import (
    click_button, get_property_proxy, set_proxy_value, singletons)

# Allows for one or two digits for day, month, hour, minute, second
# and exactly three digits for milliseconds
pattern = r"\[\d{2}-\d{2}-\d{4} \d{2}:\d{2}:\d{2}\.\d{3}\]"


class Object(Configurable):
    prop = String()
    vector = VectorString()
    integer = UInt32(
        displayType="hex|",
        defaultValue=13
    )


def test_historic_text(gui_app, subtests, mocker):
    """Test the historic text widget with string data"""
    schema = Object.getClassSchema()
    proxy = get_property_proxy(schema, "prop")
    controller = DisplayHistoricText(proxy=proxy,
                                     model=HistoricTextModel())
    controller.create(None)
    assert controller.widget is not None

    if not IS_WINDOWS_SYSTEM:
        # Note: PlainText not working on windows
        model = controller.list_model
        model.setStringList([])
        assert model.rowCount() == 0
        set_proxy_value(proxy, "prop", "Line 1")
        assert "Line 1" in controller.text_widget.toPlainText()
        set_proxy_value(proxy, "prop", "Line 2")
        assert re.search(pattern, controller.text_widget.toPlainText())
        assert "Line 2" in controller.text_widget.toPlainText()

    with subtests.test("Historic data arrival"):
        xfel = Hash("v", "xfel")
        ts_xfel = Timestamp()
        ts_xfel.toHashAttributes(xfel)
        karabo = Hash("v", "karabo")
        ts_karabo = Timestamp()
        ts_karabo.toHashAttributes(karabo)
        controller._historic_data_arrival([xfel, karabo])
        model = controller.list_model
        assert model.rowCount() == 2
        index = model.index(0, 0)
        assert re.search(pattern, index.data())

    with subtests.test("Historic data request"):
        network = mocker.Mock()
        with singletons(network=network):
            click_button(controller.request_button)
            text = controller.status_widget.text()
            assert "Requesting historic data!" in text
            network.onGetPropertyHistory.assert_called_once()

    controller.destroy()
    assert controller.widget is None


def test_vector_historic(gui_app, subtests, mocker):
    """Test the historic text widget with vector string data"""
    schema = Object.getClassSchema()
    proxy = get_property_proxy(schema, "vector")
    controller = DisplayHistoricText(proxy=proxy,
                                     model=HistoricTextModel())
    controller.create(None)
    assert controller.widget is not None

    if not IS_WINDOWS_SYSTEM:
        # Note: PlainText not working on windows
        model = controller.list_model
        model.setStringList([])
        assert model.rowCount() == 0
        set_proxy_value(proxy, "vector", ["one", "two"])
        assert "['one', 'two']" in controller.text_widget.toPlainText()
        set_proxy_value(proxy, "vector", ["one", "three", "five"])
        text = controller.text_widget.toPlainText()
        assert re.search(pattern, controller.text_widget.toPlainText())
        assert "['one', 'three', 'five']" in text

    with subtests.test("Historic data arrival"):
        first = Hash("v", ["a1", "b2", "c3"])
        ts_first = Timestamp()
        ts_first.toHashAttributes(first)
        second = Hash("v", ["one only"])
        ts_second = Timestamp()
        ts_second.toHashAttributes(second)
        third = Hash("v", [])
        ts_third = Timestamp()
        ts_third.toHashAttributes(third)
        controller._historic_data_arrival([first, second, third])
        model = controller.list_model
        assert model.rowCount() == 3
        index = model.index(0, 0)
        assert re.search(pattern, index.data())

    with subtests.test("Historic data request"):
        network = mocker.Mock()
        with singletons(network=network):
            click_button(controller.request_button)
            text = controller.status_widget.text()
            assert "Requesting historic data!" in text
            network.onGetPropertyHistory.assert_called_once()

    controller.destroy()
    assert controller.widget is None


def test_integer_hex(gui_app, subtests, mocker):
    """Test the historic text widget with vector string data"""
    schema = Object.getClassSchema()
    proxy = get_property_proxy(schema, "integer")
    controller = DisplayHistoricText(proxy=proxy,
                                     model=HistoricTextModel())
    controller.create(None)
    assert controller.widget is not None

    if not IS_WINDOWS_SYSTEM:
        # Note: PlainText not working on windows
        model = controller.list_model
        model.setStringList([])
        assert model.rowCount() == 0
        set_proxy_value(proxy, "integer", 10)
        assert "0xA" in controller.text_widget.toPlainText()
        set_proxy_value(proxy, "integer", 7)
        text = controller.text_widget.toPlainText()
        assert "0x7" in text

    with subtests.test("Historic data arrival"):
        first = Hash("v", 2)
        ts_first = Timestamp()
        ts_first.toHashAttributes(first)
        second = Hash("v", 1)
        ts_second = Timestamp()
        ts_second.toHashAttributes(second)
        third = Hash("v", 0)
        ts_third = Timestamp()
        ts_third.toHashAttributes(third)
        controller._historic_data_arrival([first, second, third])
        model = controller.list_model
        assert model.rowCount() == 3

    with subtests.test("Historic data request"):
        network = mocker.Mock()
        with singletons(network=network):
            click_button(controller.request_button)
            text = controller.status_widget.text()
            assert "Requesting historic data!" in text
            network.onGetPropertyHistory.assert_called_once()

    controller.destroy()
    assert controller.widget is None



from qtpy.QtWidgets import QDialog, QToolButton

from karabo.native import Configurable, VectorDouble
from karabogui.testing import click_button, get_property_proxy

from ..widgets import TableVectorEdit


class TableSchema(Configurable):
    vector = VectorDouble(
        displayedName="VectorDouble",
        displayType="TableVector",
        defaultValue=[])


def test_table_vector_edit(gui_app, mocker):
    proxy = get_property_proxy(TableSchema.getClassSchema(), "vector")
    widget = TableVectorEdit(proxy.binding)
    assert widget is not None
    widget.setText("2.1,2")
    assert widget.list_edit.values == ["2.1", "2"]
    assert widget.list_edit.string_values == "2.1,2"

    # Set a new value
    dialog_instance = mocker.Mock()
    dialog_instance.exec.return_value = QDialog.Accepted
    dialog_instance.string_values = "4.1"

    mocker.patch.object(widget, "list_edit", new=dialog_instance)
    button = widget.layout().itemAt(1).widget()
    assert isinstance(button, QToolButton)
    click_button(button)
    dialog_instance.exec.assert_called_once()
    assert widget.text() == "4.1"



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import inspect
import os.path as op
import pdb
from functools import wraps
from time import perf_counter

from qtpy.QtCore import pyqtRemoveInputHook

from karabogui.util import process_qt_events


def set_trace():
    """Set a breakpoint in the Python debugger that works with PyQt.
    """
    pyqtRemoveInputHook()
    pdb.set_trace()


def print_stack_trace(num_frames=0):
    """Print a stack trace, without requiring an exception to have been thrown.
    """
    frames = inspect.stack()
    IGNORED = 2  # The frames including this function and inspect.stack()
    num_frames = num_frames if num_frames != 0 else (len(frames) - IGNORED)
    frames_subset = frames[IGNORED:IGNORED + num_frames]
    try:
        print()
        for tup in frames_subset[::-1]:
            fname, line, function, context = tup[1:-1]
            print('{}:{}:{}\n{}'.format(op.basename(fname), function, line,
                                        context[0].strip()))
    finally:
        del frames
        del frames_subset


class profiler:
    """We now call profiler with arguments because functions hooked to
       paintEvents cannot be timed with invoked processEvents.
       Possible arguments are `name`, `process_events` and `timeout` for the
       event chain. The `timeout` argument is only used if the `process_events`
       setting is True (default).

       @profiler()
       def foo():
           ...

       @profiler(process_events=False, timeout=100)
       def paintEvent():
           ...

       with profiler(name="Image")
           # compute image

    """

    def __init__(self, name=None, process_events=True, timeout=200):
        self.name = name
        self.t_start = None
        self.process_events = process_events
        self.timeout = timeout

    def __enter__(self):
        if self.process_events:
            process_qt_events(timeout=self.timeout)
        self.t_start = perf_counter()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.process_events:
            process_qt_events(timeout=self.timeout)

        elapsed = perf_counter() - self.t_start
        name = f"Block {self.name}: " if self.name is not None else ""
        print(f"{name}time elapsed {elapsed}")

    def __call__(self, func):
        """Decorate a function to profile the execution time"""
        name = func.__name__ if self.name is None else self.name

        @wraps(func)
        def wrapper(*args, **kwargs):
            if self.process_events:
                process_qt_events(timeout=self.timeout)

            t_start = perf_counter()
            ret = func(*args, **kwargs)

            # But process the generated eventloop stack for the performance
            # measurement!
            if self.process_events:
                process_qt_events(timeout=self.timeout)

            elapsed = perf_counter() - t_start
            print(f"{name} took {elapsed}")
            return ret

        return wrapper



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np


def set_roi_html(name, center=None, size=None):
    html_list = []

    # Name
    html_list.append(
        f'<span style="color: #FFF; font-size: 8pt; font-weight: bold;">'
        f'{name or "Region of Interest"}</span>')
    # Center
    if center is not None:
        html_list.append(
            f'<span style="color: #FFF; font-size: 8pt;">'
            f'Center: ({center[0]}, {center[1]})</span>')
    # Size
    if size is not None:
        html_list.append(
            f'<span style="color: #FFF; font-size: 8pt;">'
            f'Size: ({size[0]}, {size[1]})</span>')

    html = "<br>".join(html_list)
    return f'<div style="text-align: center">{html}</div>'


class ImageRegion:
    Point = 0
    Line = 1
    Area = 2

    def __init__(self, region=None, region_type=None,
                 x_slice=None, y_slice=None):

        if region is None:
            region = np.empty((0, 0))
        if x_slice is None:
            x_slice = np.empty(0)
        if y_slice is None:
            y_slice = np.empty(0)

        self.region = region
        self.region_type = region_type
        self.slices = [x_slice, y_slice]

    def valid(self, axis=None):
        """Checks if region (all axis) is valid"""
        if self.region_type is ImageRegion.Area:
            # Region is a matrix, so we only check the shape
            return all(shape > 1 for shape in self.region.shape)

        if self.region_type is ImageRegion.Line:
            # Region is a list of two lists
            samples = (0, 1) if axis is None else (axis,)
            return all(self.region[ax].size > 1 for ax in samples)

        return False

    def flatten(self):
        if self.region_type is ImageRegion.Area:
            # region is a numpy array
            return self.region.flatten()

        if self.region_type is ImageRegion.Line:
            # region is a list of 2 lists (x and y most probably have different
            # shapes)
            return np.hstack(self.region)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.

from karabogui.testing import system_hash
from karabogui.topology.api import SystemTopology

from ..utils import compare_proxy_essential


def test_proxy_essential_compare():
    topology = SystemTopology()
    topology.initialize(system_hash())
    proxy = topology.get_project_device_proxy("divvy", "swerver", "FooClass")
    showing_proxy = topology.get_project_device_proxy("divvy", "swerver",
                                                      "FooClass")
    assert compare_proxy_essential(proxy, showing_proxy)

    # Project device and online device
    showing_proxy = topology.get_device("divvy", request=False)
    assert compare_proxy_essential(proxy, showing_proxy)

    # Different deviceId to compare
    showing_proxy = topology.get_device("divvyyyyy", request=False)
    assert not compare_proxy_essential(proxy, showing_proxy)

    # Two different serverids
    showing_proxy = topology.get_project_device_proxy("divvy", "swwww",
                                                      "FooClass")
    assert not compare_proxy_essential(proxy, showing_proxy)

    # Different classIds
    showing_proxy = topology.get_project_device_proxy("divvy", "swerver",
                                                      "BarClass")
    assert compare_proxy_essential(proxy, showing_proxy)



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on January 26, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from collections import defaultdict
from functools import partial

from qtpy.QtWidgets import QAction, QDialog, QMenu
from traits.api import Instance, List, on_trait_change

import karabogui.icons as icons
from karabo.common.api import walk_traits_object
from karabo.common.project.api import (
    DeviceInstanceModel, ProjectModel, device_instance_exists)
from karabo.common.project.utils import get_project_models
from karabogui import messagebox
from karabogui.access import AccessRole, access_role_allowed
from karabogui.project.dialog.project_handle import NewProjectDialog
from karabogui.project.loading_watcher import ProjectLoadingWatcher
from karabogui.project.utils import load_project
from karabogui.singletons.api import get_project_model
from karabogui.util import create_list_string, move_to_cursor

from .project_groups import ProjectSubgroupController


class SubprojectController(ProjectSubgroupController):
    """ A controller for 'subprojects' subgroups.
    """
    # These are subprojects which are asynchronously loading
    _watchers = List(Instance(ProjectLoadingWatcher))

    def context_menu(self, project_controller, parent=None):
        project_allowed = access_role_allowed(AccessRole.PROJECT_EDIT)

        menu = QMenu(parent)
        add_action = QAction(icons.add, 'Add new project...', menu)
        add_action.triggered.connect(partial(self._add_project, parent=parent))
        load_action = QAction(icons.load, 'Load project...', menu)
        load_action.triggered.connect(partial(self._load_project,
                                              parent=parent))
        add_action.setEnabled(project_allowed)
        load_action.setEnabled(project_allowed)

        menu.addAction(add_action)
        menu.addAction(load_action)
        return menu

    # ----------------------------------------------------------------------
    # action handlers

    def _add_project(self, parent=None):
        """ Add a new subproject to the associated project
        """
        dialog = NewProjectDialog(default=True, parent=parent)
        move_to_cursor(dialog)
        if dialog.exec() == QDialog.Accepted:
            # XXX: TODO check for existing
            project = ProjectModel(simple_name=dialog.simple_name)
            # Set initialized and modified last
            project.initialized = project.modified = True
            self.model.subprojects.append(project)

    def _load_project(self, parent=None):
        """ Add an existing project as a subproject.
        """
        project = load_project(is_subproject=True, parent=parent)
        if project is None:
            return

        # Loading projects is complex, due to asynchronicity. Use a helper
        # to determine when all the project data is present.
        watcher = ProjectLoadingWatcher(project=project)
        if watcher.finished:
            self._check_and_add_project(project)
        else:
            self._watchers.append(watcher)

    # ----------------------------------------------------------------------
    # trait handlers

    @on_trait_change('_watchers:finished_event')
    def _watcher_finished(self, obj, name, old, new):
        """Called when a project finishes loading all of its data
        """
        self._watchers.remove(obj)
        self._check_and_add_project(obj.project)

    # ----------------------------------------------------------------------
    # util methods

    def _check_and_add_project(self, project):
        """Make sure a subproject doesn't have conflicts with the open project
        """
        root_project = get_project_model().root_model
        if self._check_for_duplicate_projects(project, root_project):
            return
        if self._check_for_duplicate_devices(project, root_project):
            return

        # Finally add the subproject
        self.model.subprojects.append(project)

    def _check_for_duplicate_devices(self, project, root_project):
        """Make sure a subproject doesn't have conflicting device IDs
        """
        device_ids = []
        duplicated = []

        def visitor(model):
            if isinstance(model, DeviceInstanceModel):
                device_ids.append(model.instance_id)

        # collect all the device ids in the project being added
        walk_traits_object(project, visitor)

        for project in get_project_models(root_project):
            if devices := device_instance_exists(project, device_ids):
                duplicated.extend(devices)

        if duplicated:
            msg = ('The subproject which you wish to load contains devices<br>'
                   'which already exist in the current project. Therefore <br>'
                   'it will not be added.')

            details = create_list_string(set(duplicated))
            messagebox.show_warning(msg, details=details,
                                    title='Device already exists')
            return True
        return False

    def _check_for_duplicate_projects(self, project, root_project):
        """Make sure a subproject doesn't include any projects which are
        already loaded.
        """
        project_uuids = defaultdict(int)
        project_names = []
        empty_projects = 0

        def _visitor(model):
            """Count the number of times each UUID appears in the tree"""
            nonlocal project_uuids, project_names, empty_projects
            if isinstance(model, ProjectModel):
                project_uuids[model.uuid] += 1
                if not model.simple_name:
                    empty_projects += 1
                if project_uuids[model.uuid] > 1:
                    if model.simple_name:
                        project_names.append(model.simple_name)

        # Walk both projects and count the instances of project models
        walk_traits_object(project, _visitor)
        walk_traits_object(root_project, _visitor)

        # If a UUID appeared more than once, we have a problem!
        if any(v > 1 for v in project_uuids.values()):
            msg = ('Projects will be OR are loaded already multiple times! '
                   'Please investigate and have a look at the '
                   'project(s) <b>{}</b>! You have <b>{}</b> projects '
                   'that are loaded with an empty name!'.format(
                    project_names, empty_projects))
            messagebox.show_warning(msg, title='Load Subproject Failed')
            return True
        return False



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
import pytest

from karabo.common.scenemodel.api import EvaluatorModel
from karabo.native import Configurable, Float
from karabogui.binding.api import (
    DeviceClassProxy, PropertyProxy, ProxyStatus, build_binding)
from karabogui.controllers.display.eval import Evaluator
from karabogui.indicators import ALL_OK_COLOR
from karabogui.testing import set_proxy_value


class Object(Configurable):
    prop = Float()
    alarms = Float()


@pytest.fixture
def evaluator_setup(gui_app):
    schema = Object.getClassSchema()
    binding = build_binding(schema)
    device = DeviceClassProxy(binding=binding, server_id="Fake",
                              status=ProxyStatus.OFFLINE)
    prop = PropertyProxy(root_proxy=device, path="prop")
    alarms = PropertyProxy(root_proxy=device, path="alarms")
    yield prop, alarms


def test_basics(evaluator_setup):
    prop, _ = evaluator_setup
    controller = Evaluator(proxy=prop, model=EvaluatorModel())
    controller.create(None)
    assert controller.widget is not None

    controller.destroy()
    assert controller.widget is None


def test_set_value(evaluator_setup):
    prop, _ = evaluator_setup
    model = EvaluatorModel(expression="x**2")
    controller = Evaluator(proxy=prop, model=model)
    controller.create(None)
    set_proxy_value(prop, "prop", 2.0)
    assert controller._internal_widget.text() == "4.0"


def test_builtin_function(evaluator_setup):
    prop, _ = evaluator_setup
    model = EvaluatorModel(expression="round(x)")
    controller = Evaluator(proxy=prop, model=model)
    controller.create(None)
    set_proxy_value(prop, "prop", 2.5)
    assert controller._internal_widget.text() == "2.0"


def test_change_expression(evaluator_setup, mocker):
    prop, _ = evaluator_setup
    set_proxy_value(prop, "prop", 21.0)
    controller = Evaluator(proxy=prop, model=EvaluatorModel())
    controller.create(None)
    action = controller.widget.actions()[0]
    assert action.text() == "Change expression..."

    dsym = "karabogui.controllers.display.eval.QInputDialog"
    QInputDialog = mocker.patch(dsym)
    QInputDialog.getText.return_value = "x*2", True
    action.trigger()
    assert controller._internal_widget.text() == "42.0"

    # Cause a SyntaxError messagebox
    msym = "karabogui.controllers.display.eval.messagebox"
    messagebox = mocker.patch(msym)
    QInputDialog.getText.return_value = "sin(x", True
    action.trigger()
    assert messagebox.show_warning.call_count == 1


def test_alarm_warning_fine_color(evaluator_setup):
    _, alarms = evaluator_setup
    controller = Evaluator(proxy=alarms)
    controller.create(None)

    set_proxy_value(alarms, "alarms", 0.75)
    assert controller._internal_widget.text() == "0.75"
    assert controller._bg_color == ALL_OK_COLOR

    set_proxy_value(alarms, "alarms", 3.0)
    assert controller._bg_color == ALL_OK_COLOR

    set_proxy_value(alarms, "alarms", 1.5)
    assert controller._bg_color == ALL_OK_COLOR


def test_type_conversion(evaluator_setup):
    prop, _ = evaluator_setup
    model = EvaluatorModel(expression="round(x,6)")
    controller = Evaluator(proxy=prop, model=model)
    controller.create(None)
    set_proxy_value(prop, "prop", np.float32(2.1234567899999))
    assert controller._internal_widget.text() == "2.123457"



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import datetime
import time
from functools import partial

import numpy as np
from qtpy.QtCore import QPoint, Qt
from qtpy.QtGui import QColor, QPainter, QPainterPath, QPen
from qtpy.QtWidgets import (
    QAction, QActionGroup, QHBoxLayout, QInputDialog, QLineEdit, QWidget)
from traits.api import Instance, on_trait_change

from karabo.common.scenemodel.api import SparklineModel
from karabo.native import Timestamp
from karabogui.binding.api import FloatBinding, IntBinding, get_binding_value
from karabogui.controllers.api import (
    BaseBindingController, register_binding_controller)
from karabogui.util import SignalBlocker

SPARK_MIN_HEIGHT = 50
SPARK_MIN_WIDTH = 200
# timebases available for the widget, values in seconds
TIMEBASES = (('60s', 60), ('10m', 600), ('10h', 36000))


class SparkRenderer(QWidget):
    """A renderer for sparklines using QPainter
    """
    # the widget has a fixed time base
    time_base = 600  # seconds
    # maximum number of points to use, determines the binning of data
    _p_max = 120
    # plot fraction in y-direction to use
    _plot_frac = 0.9
    # size in pixels of the min max indication area
    _min_max_indic_size = 60
    # relative epsilon within within the tendency will be considered stable
    _tendency_eps = 0.05
    # size of the window in bins which is used for evaluating trend indication
    _indicator_window_size = 5

    def __init__(self, parent):
        super().__init__(parent)

        self.pen = QPen(Qt.black, 2, Qt.SolidLine)
        self.no_pen = QPen(Qt.blue, 1, Qt.SolidLine)

        self.yrange = None
        self.reset_range = False
        self.ystep = None
        self.tvals = np.arange(self._p_max)

        self.yvals = np.zeros(self._p_max, np.double)

        self.finfo = np.finfo(np.float64)
        self.ymax = np.ones(self._p_max, np.float64) * self.finfo.min
        self.ymin = np.ones(self._p_max, np.float64) * self.finfo.max
        self.ycnts = np.zeros(self._p_max, int)

        self.painter_path = None
        self.painter_path_min = None
        self.painter_path_max = None

        self.then = time.time()

        self.tendency = 0
        self.tendency_indicators = {}
        self.tendency_indicators[0] = {'x': np.array([0., 10., 0., 0.]),
                                       'y': np.array([-5., 0., 5., -5.])}

        self.tendency_indicators[1] = {'x': np.array([-5., 0., 5., -5.]),
                                       'y': np.array([10., 0., 10., 10.])}

        self.tendency_indicators[-1] = {'x': np.array([-5., 0., 5., -5.]),
                                        'y': np.array([-7., 3., -7., -7.])}

    def _scale_y(self, val, height):
        if self.yrange is None:
            return

        dy = self.yrange[1] - self.yrange[0]
        if dy == 0:
            return

        scaled = (val - self.yrange[0]) / dy
        scaled /= self.ystep
        scaled = (height - scaled + self.offset).astype(np.int64)
        return scaled

    def resizeEvent(self, event):
        self._update_ranges(self.yvals[-1])
        self._generate_curve()
        event.accept()

    def _generate_curve(self):
        width = self.width() - self._min_max_indic_size
        height = self._plot_frac * self.height()
        step = width / self._p_max
        nonzero = np.nonzero(self.ycnts)
        x = (self.tvals[nonzero] * step).astype(np.int32)
        y = self._scale_y(self.yvals[nonzero] / self.ycnts[nonzero], height)

        if y is None:
            return

        ymin = self._scale_y(self.ymin[nonzero], height)
        ymax = self._scale_y(self.ymax[nonzero], height)

        # evaluate tendency for last 5 point intervals
        iws = self._indicator_window_size
        c_idx_nz = np.nonzero(self.ycnts[-iws:])
        t_idx_nz = np.nonzero(self.ycnts[-2 * iws:-iws])
        if np.any(c_idx_nz) and np.any(t_idx_nz):
            current_avg = np.mean(self.yvals[-iws:][c_idx_nz]
                                  / self.ycnts[-iws:][c_idx_nz])
            test_avg = np.mean(self.yvals[-2 * iws:-iws][t_idx_nz]
                               / self.ycnts[-2 * iws:-iws][t_idx_nz])
            delta = (current_avg - test_avg) / test_avg if test_avg else 0.
            if abs(delta) <= self._tendency_eps or not np.isfinite(delta):
                self.tendency = 0
            else:
                self.tendency = np.sign(delta)

        self.painter_path = self._createPainterPath(x, y)
        self.painter_path_min = self._createPainterPath(x, ymin)
        self.painter_path_max = self._createPainterPath(x, ymax)

    def paintEvent(self, event):
        if self.painter_path is None:
            return

        width = self.width() - self._min_max_indic_size
        height = self._plot_frac * self.height()

        with QPainter(self) as painter:
            painter.setPen(self.pen)

            painter.setPen(self.pen)
            painter.drawPath(self.painter_path)

            pen = QPen(Qt.gray, 1, Qt.SolidLine)
            painter.setPen(pen)

            painter.drawPath(self.painter_path_min)
            painter.drawPath(self.painter_path_max)

            # add min and max value indicators
            pen = QPen(Qt.black, 1, Qt.SolidLine)
            painter.setPen(pen)
            x_points = [width, width]
            y_points = [0, self.height()]
            painter.drawPath(self._createPainterPath(np.array(x_points),
                                                     np.array(y_points)))

            # add min max indicator
            font = painter.font()
            font.setPointSize(8)
            font.setFamily("Monospace")
            painter.setFont(font)
            painter.drawText(QPoint(int(width) + 5, 10),
                             self._returnFormatted(self.yrange[1]))
            painter.drawText(QPoint(int(width) + 5, int(height) + 2),
                             self._returnFormatted(self.yrange[0]))

            #  add tendency indicator
            x_points = self.tendency_indicators[self.tendency]['x'] + width
            y_points = self.tendency_indicators[self.tendency]['y']
            dy = height / 2 + 2.5
            if self.tendency != 0:
                dy = height / 2 + (-1) * self.tendency * height / 2
            path = self._createPainterPath(np.array(x_points),
                                           np.array(y_points) + dy)
            painter.fillPath(path, Qt.magenta)
            painter.drawPath(path)

            # add timebase info
            tb = self.time_base
            if tb >= 3600:
                tb = f"{tb // 3600:d}h"
            elif tb >= 600:
                tb = f"{tb // 60:d}m"
            else:
                tb = f"{tb:d}s"

            dy = self.height() - 13
            path = self._createPainterPath(np.array([3., 28., 28., 3., 3.]),
                                           np.array(
                                               [0., 0., 12., 12., 0.]) + dy)
            painter.fillPath(path, QColor(255, 255, 255, 200))
            painter.drawPath(path)
            painter.drawText(QPoint(5, self.height() - 2), tb)

    def _returnFormatted(self, v):
        if v == 0:
            return f"{v:+0.2f}"

        exp = np.log10(abs(v))
        if abs(exp) <= 2:
            return f"{v:+0.2f}"
        else:
            return f"{v:+0.1e}"

    def _createPainterPath(self, x, y):
        path = QPainterPath()
        path.moveTo(x[0], y[0])
        for i in range(1, x.size):
            path.lineTo(x[i], y[i])
        return path

    def setData(self, value, timestamp):
        """Set a new data point with associated timestamp to the spark line
        Set the value to `None` if no bound is set.
        """
        t = timestamp.toTimestamp()

        # check if we need to roll to the next bin
        dt = int((t - self.then) / (self.time_base / self._p_max))
        if dt >= 1:
            # the sparkline might need to cover "gaps" in its data series
            # e.g. when for a given bin, no data was pushed to the widget
            # in this case we fill with the value of the previous bin.
            # If the sparkline does not receive any value updates from a
            # device for dt > time_base, indexing by [-dt:] would result
            # in an IndexError. In these cases we set all values in the
            # sparkline to the value of the last update, consitent with the
            # gap filling behaviour for valid "dt"s.
            if dt < self.yvals.size:
                self.yvals = np.roll(self.yvals, -dt)
                self.yvals[-dt:-1] = self.yvals[-dt - 1]
                self.yvals[-1] = 0.
                self.ycnts = np.roll(self.ycnts, -dt)
                self.ycnts[-dt:-1] = self.ycnts[-dt - 1]
                self.ycnts[-1] = 0.
                self.ymax = np.roll(self.ymax, -dt)
                self.ymax[-dt:-1] = self.ymax[-dt - 1]
                self.ymin = np.roll(self.ymin, -dt)
                self.ymin[-dt:-1] = self.ymin[-dt - 1]
            else:
                self.yvals[:] = self.yvals[-1]
                self.ycnts[:] = self.ycnts[-1]
                self.ymax[:] = self.ymax[-1]
                self.ymin[:] = self.ymin[-1]

            self.then = time.time()

        self.yvals[-1] += value
        self.ycnts[-1] += 1
        self.ymax[-1] = max(self.ymax[-1], value)
        self.ymin[-1] = min(self.ymin[-1], value)
        self._update_ranges(value)
        # generate the curve with the painter path
        self._generate_curve()

    def setSeries(self, x, y):
        """Set a series of data to the sparkline, e.g. historic data

        Note that this will not redraw the sparkline until the next
        setData update.
        """
        assert x.size == y.size

        if y.size == 0:
            return

        now = time.time()
        x = now - x
        binned, bins = np.histogram(x, bins=self._p_max,
                                    range=[0, self.time_base], weights=y)

        counts, _ = np.histogram(x, bins=self._p_max,
                                 range=[0, self.time_base])

        self.yvals = binned[::-1]
        self.ycnts = counts[::-1]

        bin_idx = np.digitize(x, bins)

        self.ymin = np.zeros(self.yvals.shape, np.float64)
        self.ymax = np.zeros(self.yvals.shape, np.float64)

        for i in range(self._p_max):
            idx = bin_idx == i
            if idx.any():
                self.ymin[i] = np.min(y[idx])
                self.ymax[i] = np.max(y[idx])

        self.ymin = self.ymin[::-1]
        self.ymax = self.ymax[::-1]

        self.yrange = (np.min(y), np.max(y))

    def _update_ranges(self, value):
        """Update data ranges as needed, depending on values.
        """
        # the y range which simply needs to scale to new maxima and minima
        if self.yrange is None or self.reset_range:
            # initially only single value
            self.yrange = (min(0.9 * value, 1.1 * value),
                           max(0.9 * value, 1.1 * value))
            self.reset_range = False
        else:
            self.yrange = min(value, *self.yrange), max(value, *self.yrange)

        # scale to plot fraction
        pheight = self._plot_frac * self.height()
        self.offset = (self.height() - pheight) // 2
        self.ystep = 1. / pheight

    def update_time_base(self, time_base):
        self.time_base = time_base


@register_binding_controller(ui_name='Sparkline', klassname='DisplaySparkline',
                             binding_type=(FloatBinding, IntBinding),
                             can_show_nothing=False)
class DisplaySparkline(BaseBindingController):
    """Displays a spark line with a fixed time basis and coarsing of data

    On initialization and visibility changes historic data is requested.
    Rendering is performed via the SparkRenderer class
    """
    # The scene model class used by this controller
    model = Instance(SparklineModel, args=())

    line_edit = Instance(QLineEdit)
    render_area = Instance(SparkRenderer)
    action_show_format = Instance(QAction)

    def create_widget(self, parent):
        widget = QWidget(parent)
        layout = QHBoxLayout(widget)
        widget.setMinimumHeight(SPARK_MIN_HEIGHT)

        self.line_edit = QLineEdit(widget)
        self.line_edit.setReadOnly(True)
        self.render_area = SparkRenderer(widget)
        self.render_area.setMinimumWidth(SPARK_MIN_WIDTH)
        self.render_area.time_base = self.model.time_base

        layout.addWidget(self.line_edit)
        self.line_edit.setVisible(self.model.show_value)
        layout.addWidget(self.render_area)

        action_show_value = QAction("Show value", widget, checkable=True)
        action_show_value.toggled.connect(self.toggle_show_value)
        action_show_value.setChecked(self.model.show_value)
        widget.addAction(action_show_value)

        self.action_show_format = QAction("Set value format", widget)
        self.action_show_format.triggered.connect(self._change_show_format)
        self.action_show_format.setEnabled(self.model.show_value)
        widget.addAction(self.action_show_format)

        def change_time_base(tb):
            self.model.time_base = tb

        time_base_separator = QAction("Timebase", widget)
        time_base_separator.setSeparator(True)
        widget.addAction(time_base_separator)
        time_base_group = QActionGroup(widget)
        for label, base in TIMEBASES:
            tbAction = QAction(f"{label} time base",
                               time_base_group, checkable=True)
            tbAction.toggled.connect(partial(change_time_base, base))

            if self.model.time_base == base:
                tbAction.setChecked(True)

            time_base_group.addAction(tbAction)

        widget.addActions(time_base_group.actions())
        return widget

    def binding_update(self, proxy):
        now = time.time()
        self._fetch_property_history(now - self.model.time_base, now)

    def value_update(self, proxy):
        self._draw(proxy.value, proxy.binding.timestamp)

    def toggle_show_value(self, value):
        self.model.show_value = value

    def toggle_alarm_range(self, value):
        self.model.alarm_range = value

    @on_trait_change('proxy:binding:historic_data')
    def _historic_data_arrived(self, data):
        if not data:
            return

        datasize = len(data)
        x = np.empty(datasize, dtype=float)
        y = np.empty(datasize, dtype=float)

        for i, d in enumerate(data):
            x[i] = Timestamp.fromHashAttributes(d['v', ...]).toTimestamp()
            y[i] = d["v"]

        self.render_area.setSeries(x[:-1], y[:-1])
        # use the last data point to trigger a redraw
        self._draw(y[-1], Timestamp.fromHashAttributes(data[-1]['v', ...]))

    @on_trait_change('proxy.visible', post_init=True)
    def _visibility_update(self, visible):
        if visible:
            now = time.time()
            self._fetch_property_history(now - self.model.time_base, now)

    @on_trait_change('model:time_base')
    def _update_time_base(self, timebase):
        self.render_area.update_time_base(timebase)
        now = time.time()
        self._fetch_property_history(now - timebase, now)

    @on_trait_change('model:show_value')
    def _show_value_update(self, value):
        self.line_edit.setVisible(value)
        self.action_show_format.setEnabled(value)

    @on_trait_change('model:show_format')
    def _show_format_update(self):
        self._set_text()

    def _change_show_format(self):
        form, ok = QInputDialog.getText(self.widget, "Enter Format",
                                        "", text=self.model.show_format)
        if ok:
            self.model.show_format = form

    def _draw(self, value, timestamp):
        """ Draw data vs. time"""
        self.render_area.setData(value, timestamp)
        self._set_text()

    def _fetch_property_history(self, t0, t1):
        if not self.proxy.visible:
            return
        t0 = str(datetime.datetime.fromtimestamp(t0,
                                                 tz=datetime.UTC).isoformat())
        t1 = str(datetime.datetime.fromtimestamp(t1,
                                                 tz=datetime.UTC).isoformat())
        self.proxy.get_history(t0, t1)

    def _set_text(self):
        value = get_binding_value(self.proxy)
        if self.line_edit is None or value is None:
            return
        template = "{:" + self.model.show_format + "}"
        with SignalBlocker(self.line_edit):
            self.line_edit.setText(template.format(value))



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import unittest

from karabogui.graph.plots.utils import _get_sample_threshold

TYPICAL = 20000


class TestPlotUtils(unittest.TestCase):

    def setUp(self):
        super().setUp()

    def tearDown(self):
        super().tearDown()

    def test_sample_threshold(self):
        """Test the sample threshold finding"""
        threshold = _get_sample_threshold(1000)
        assert threshold == TYPICAL
        threshold = _get_sample_threshold(10000)
        assert threshold == TYPICAL
        threshold = _get_sample_threshold(200000)
        assert threshold == TYPICAL
        threshold = _get_sample_threshold(200001)
        assert threshold == 30000
        threshold = _get_sample_threshold(300000)
        assert threshold == 30000
        threshold = _get_sample_threshold(600000)
        assert threshold == 60000



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on January 12, 2017
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtCore import QItemSelectionModel, QObject


class SelectionTracker(QObject):
    """An object which keeps track of what is selected.
    """

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self._model = None

    def grab_selection(self, selection_model):
        """'Grabs' the current selection.

        In this way, multiple selection models can maintain mutually exclusive
        selection with each other. Whenever one model gets a selection, it
        calls this method and the previously selected model is cleared. Then
        it becomes the currently selected model until this method is called
        with a different selection model.
        """
        assert isinstance(selection_model, QItemSelectionModel)

        if self._model is not None and self._model is not selection_model:
            self._model.clear()
        self._model = selection_model



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtCore import QPoint, QRect, Qt
from qtpy.QtGui import QColor, QPen
from traits.api import Any, Enum, HasStrictTraits, Instance, String

from karabogui.binding.api import DeviceProxy
from karabogui.events import KaraboEvent, broadcast_event
from karabogui.sceneview.bases import BaseSceneTool
from karabogui.sceneview.const import HOVER_COLOR
from karabogui.sceneview.layout.layouts import GroupLayout
from karabogui.sceneview.utils import (
    round_down_to_grid, round_up_to_grid, save_painter_state)
from karabogui.sceneview.widget.unknown import UnknownSvgWidget

NONRESIZABLE_OBJECTS = (GroupLayout,)
NONSELECTABLE_OBJECTS = (UnknownSvgWidget,)


class ProxySelectionTool(BaseSceneTool):
    """A tool for the selection of widgets in the SceneView which are derived
    from ``ControllerContainer`` and bound to proxies.
    """

    def mouse_down(self, scene_view, event):
        # Only controller might have proxies
        controller = scene_view.controller_at_position(event.pos())
        if controller is not None:
            # Use the main proxy of the controller
            proxy = controller.widget_controller.proxy
            device = proxy.root_proxy
            if isinstance(device, DeviceProxy):  # ignore DeviceClassProxy
                if not device.online:
                    return  # ignore offline devices
                broadcast_event(KaraboEvent.ShowConfiguration,
                                {'proxy': device})
            return

    def mouse_move(self, scene_view, event):
        pass

    def mouse_up(self, scene_view, event):
        pass


class _SelectionRect(HasStrictTraits):
    selection_start = Instance(QPoint, args=())
    selection_stop = Instance(QPoint, args=())

    def start_drag(self, pos):
        self.selection_start = self.selection_stop = pos

    def rect(self):
        return QRect(self.selection_start, self.selection_stop)

    def draw(self, painter):
        """Draw the selection.
        """
        with save_painter_state(painter):
            painter.setPen(Qt.black)
            painter.drawRect(self.rect())


HOVER_TOLERANCE = 7
HOVER_QCOLOR = QColor(*(HOVER_COLOR + (128,)))


class SceneSelectionTool(BaseSceneTool):
    """A tool for selecting things in the SceneView
    """
    visible = True

    state = Enum('normal', 'draw', 'move', 'resize')

    _hover_item = Any
    _moving_pos = Instance(QPoint, args=())
    _resize_type = String
    _selection_rect = Instance(_SelectionRect, args=())

    def draw(self, scene_view, painter):
        """Draw the tool
        """
        selection_model = scene_view.selection_model
        self._draw_hovered_item(selection_model, painter)

        if self.state == 'draw':
            self._selection_rect.draw(painter)

    def mouse_down(self, scene_view, event):
        """A callback which is fired whenever the user clicks in the
        SceneView.
        """
        selection_model = scene_view.selection_model

        if self._hover_item is not None and len(self._resize_type) > 0:
            # We select the hover item
            self.state = 'resize'
            selection_model.clear_selection()
            selection_model.select_object(self._hover_item)
            event.accept()
            return

        mouse_pos = event.pos()

        if selection_model.has_multiple_selection():
            # If there is a multiple selection, we start moving it
            if selection_model.get_selection_bounds().contains(mouse_pos):
                self.state = 'move'
                self._moving_pos = mouse_pos
                event.accept()
                return

        item = scene_view.item_at_position(mouse_pos)
        if item is None:
            # No items are selected. We draw a selection rectangle instead.
            self.state = 'draw'
            self._selection_rect.start_drag(mouse_pos)
        else:
            # There is a selected or hovered item, which has preference.
            # We select and start moving it.
            item = self._hover_item or item
            self.state = 'move'
            self._moving_pos = mouse_pos
            if event.modifiers() & Qt.ShiftModifier:
                if item in selection_model:
                    selection_model.deselect_object(item)
                else:
                    selection_model.select_object(item)
            elif item not in selection_model:
                selection_model.clear_selection()
                selection_model.select_object(item)
        event.accept()

    def mouse_move(self, scene_view, event):
        """A callback which is fired whenever the user moves the mouse
        in the SceneView.
        """
        if not event.buttons():
            self._hover_move(scene_view, event)
        else:
            if self.state == 'move':
                self._move_move(scene_view, event)
            elif self.state == 'draw':
                self._draw_move(scene_view, event)
            elif self.state == 'resize':
                self._resize_move(scene_view, event)

    def mouse_up(self, scene_view, event):
        """A callback which is fired whenever the user ends a mouse click
        in the SceneView.
        """
        if self.state == 'draw':
            selection_model = scene_view.selection_model
            if not event.modifiers() & Qt.ShiftModifier:
                selection_model.clear_selection()
            items = scene_view.items_in_rect(self._selection_rect.rect())
            for item in items:
                selection_model.select_object(item)
            event.accept()

        self._hover_item = None
        self._resize_type = ''
        self.state = 'normal'

    def _draw_move(self, scene_view, event):
        """Handle mouse moves when drawing the selection rect.
        """
        self._selection_rect.selection_stop = event.pos()
        event.accept()

    def _hover_move(self, scene_view, event):
        """Handles mouse moves when no buttons are pressed.
        """
        mouse_pos = event.pos()
        selection_model = scene_view.selection_model
        cursor = 'none'
        self._resize_type = ''

        if selection_model.has_multiple_selection():
            # Multiple selections. Does not support resizing.
            # Always in move mode.
            if selection_model.get_selection_bounds().contains(mouse_pos):
                cursor = 'open-hand'
        else:
            # Single selection. We check for move or resize mode
            self._hover_item = hover_item = scene_view.item_at_hover(mouse_pos)
            if hover_item is not None:
                # The following objects do not support resizing.
                # Always in move mode
                cursor = 'open-hand'
                if is_resizable(hover_item):
                    cursor = self._get_cursor(mouse_pos)

        scene_view.set_cursor(cursor)

    def _get_cursor(self, mouse_pos):
        rect = self._hover_item.geometry()
        top = abs(rect.top() - mouse_pos.y()) < HOVER_TOLERANCE
        bottom = abs(rect.bottom() - mouse_pos.y()) < HOVER_TOLERANCE
        left = abs(rect.left() - mouse_pos.x()) < HOVER_TOLERANCE
        right = abs(rect.right() - mouse_pos.x()) < HOVER_TOLERANCE
        on_sides = left or right
        on_ends = top or bottom

        if on_ends and not on_sides:
            cursor = 'resize-vertical'
            self._resize_type = 't' if top else 'b'
        elif on_sides and not on_ends:
            cursor = 'resize-horizontal'
            self._resize_type = 'l' if left else 'r'
        elif (top and left) or (bottom and right):
            cursor = 'resize-diagonal-tlbr'
            self._resize_type = 'tl' if top else 'br'
        elif (top and right) or (bottom and left):
            cursor = 'resize-diagonal-trbl'
            self._resize_type = 'tr' if top else 'bl'
        else:
            cursor = 'open-hand'
            self._resize_type = ''
        return cursor

    def _move_move(self, scene_view, event):
        """Handles mouse moves when a scene object is being moved.
        """
        mouse_pos = event.pos()
        # Create a rect with scene view size and origin (0, 0)!
        og = QRect(QPoint(0, 0), scene_view.size())
        if not og.contains(mouse_pos):
            return
        trans = mouse_pos - self._moving_pos

        # Calculate snap translation
        if scene_view.snap_to_grid:
            origin = scene_view.selection_model.get_item_rect().topLeft()
            trans = _get_snap_offset(origin, offset=trans)

        # Translate only if there is an offset (can be zero from snap calc)
        if trans:
            for c in scene_view.selection_model:
                c.translate(trans)
            self._moving_pos += trans
        event.accept()

    def _resize_move(self, scene_view, event):
        """Handles mouse moves when a scene object is being resized.
        """
        mouse_pos = event.pos()
        og = self._hover_item.geometry()
        g = QRect(og)
        posX, posY = mouse_pos.x(), mouse_pos.y()

        # Get snap position if scene is in grid mode
        if scene_view.snap_to_grid:
            posX, posY = round_down_to_grid(posX), round_down_to_grid(posY)

        if "t" in self._resize_type:
            g.setTop(posY)
        elif "b" in self._resize_type:
            g.setBottom(posY)
        if "l" in self._resize_type:
            g.setLeft(posX)
        elif "r" in self._resize_type:
            g.setRight(posX)

        min_size = self._hover_item.minimumSize()
        max_size = self._hover_item.maximumSize()

        if (not min_size.width() <= g.size().width() <= max_size.width() or
            not min_size.height() <= g.size().height() <= max_size.height() and
            (min_size.width() <= og.size().width() <= max_size.width() and
                min_size.height() <= og.size().height() <= max_size.height())):
            return

        self._hover_item.set_geometry(g)

    def _draw_hovered_item(self, selection_model, painter):
        hovered_item = self._hover_item

        # Don't draw an indicator if there's no hovered item
        if hovered_item is None:
            return

        # Also don't draw if the single selection is the hovered item
        selected_item = selection_model.get_single_selection()
        if hovered_item is selected_item:
            return

        pen = QPen(HOVER_QCOLOR)
        pen.setWidth(3)

        with save_painter_state(painter):
            painter.setPen(pen)
            painter.drawRect(hovered_item.geometry())


def _get_snap_offset(qpoint, offset=QPoint()):
    """Calculate snap offset from on the grid step and mouse offset"""
    trans_point = qpoint + offset

    # Get rounding function
    x_round = _round_func(offset.x())
    y_round = _round_func(offset.y())

    snapped_point = QPoint(x_round(trans_point.x()), y_round(trans_point.y()))

    return snapped_point - qpoint


def _round_func(direction):
    func = round_down_to_grid
    if direction <= 0:
        func = round_up_to_grid
    return func


def is_resizable(gui_obj):
    return not isinstance(gui_obj, NONRESIZABLE_OBJECTS)


def is_selectable(gui_obj):
    return not isinstance(gui_obj, NONSELECTABLE_OBJECTS)



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on January 5, 2022
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from pathlib import Path

from qtpy.QtCore import QSize
from qtpy.QtGui import QPixmap
from qtpy.QtSvg import QSvgRenderer, QSvgWidget
from qtpy.QtWidgets import QHBoxLayout, QLabel, QWidget

from karabo.common.scenemodel.api import extract_base64image
from karabogui import icons
from karabogui.widgets.hints import KaraboSceneWidget

MISSING_ICON = str(Path(icons.__file__).parent / "image-missing.svg")


class ImageRendererWidget(KaraboSceneWidget, QWidget):

    def __init__(self, model, parent=None):
        super().__init__(model=model, parent=parent)
        layout = QHBoxLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        image_format, data = extract_base64image(model.image)
        self.setToolTip(f"ImageRenderer - Format: {image_format}")
        if image_format == "svg":
            widget = QSvgWidget(self)
            renderer = QSvgRenderer(data)
            if not renderer.isValid() or renderer.defaultSize().isNull():
                data = MISSING_ICON
            widget.load(data)
        else:
            widget = QLabel(self)
            widget.setScaledContents(True)
            pixmap = QPixmap()
            pixmap.loadFromData(data)
            if pixmap.isNull():
                pixmap = icons.imageMissing.pixmap(QSize(50, 50))
            widget.setPixmap(pixmap)

        self._internal_widget = widget
        self.layout().addWidget(widget)
        self.setGeometry(model.x, model.y, model.width, model.height)

    def add_proxies(self, proxies):
        """Satisfy the informal widget interface."""

    def apply_changes(self):
        """Satisfy the informal widget interface."""

    def decline_changes(self):
        """Satisfy the informal widget interface."""

    def destroy(self):
        """Satisfy the informal widget interface."""
        self._internal_widget.destroy()
        super().destroy()

    def set_visible(self, visible):
        """Satisfy the informal widget interface."""

    def update_global_access_level(self, level):
        """Satisfy the informal widget interface."""

    def set_geometry(self, rect):
        """Satisfy the informal widget interface."""
        self.model.trait_set(x=rect.x(), y=rect.y(),
                             width=rect.width(), height=rect.height())
        self.setGeometry(rect)

    def translate(self, offset):
        new_pos = self.pos() + offset
        self.model.trait_set(x=new_pos.x(), y=new_pos.y())
        self.move(new_pos)



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

from qtpy.QtCore import QModelIndex, Qt
from qtpy.QtWidgets import (
    QAbstractItemView, QAction, QComboBox, QDialog, QHBoxLayout, QHeaderView,
    QInputDialog, QLayout, QLineEdit, QMenu, QPushButton, QVBoxLayout, QWidget)
from traits.api import Bool, Dict, Instance, Type, Undefined, WeakRef

import karabogui.icons as icons
from karabo.common.api import KARABO_SCHEMA_MAX_SIZE, KARABO_SCHEMA_MIN_SIZE
from karabogui.binding.api import (
    StringBinding, get_default_value, get_editor_value)
from karabogui.controllers.api import (
    BaseBindingController, has_options, is_proxy_allowed)
from karabogui.dialogs.api import TopologyDeviceDialog
from karabogui.util import SignalBlocker

from .delegates import (
    TableButtonDelegate, get_display_delegate, get_table_delegate)
from .filter_model import TableSortFilterModel
from .model import TableModel
from .utils import is_writable_binding
from .view import KaraboTableView


class BaseTableController(BaseBindingController):
    """The BaseTableController that can be used for subclassing Tables

    This class provides a read only and reconfigurable version of the table
    element.
    """
    # The scene model class used by this controller
    model = Undefined
    # Does the controller have an own menu. Subclass method `custom_menu`
    # if this is set to `True`
    hasCustomMenu = Bool(False)
    # Overwrite tableModel to specify a custom type of `TableModel`
    tableModelClass = Type(TableModel)

    # Internal traits
    _bindings = Dict
    _readonly = Bool(True)
    _item_model = WeakRef(TableModel, allow_none=True)
    _table_widget = WeakRef(KaraboTableView)
    _table_buttons = Bool(False)

    _hasResize = Bool(False)

    # ---------------------------------------------------------------------
    # Abstract Methods

    def create_widget(self, parent):
        assert self.model is not Undefined

        table_widget = KaraboTableView(parent=parent)
        table_widget.setSelectionBehavior(QAbstractItemView.SelectItems
                                          | QAbstractItemView.SelectRows)
        table_widget.horizontalHeader().setStretchLastSection(True)
        # Create an internal weak ref to subclass create widget
        self._table_widget = table_widget

        # Not every table model might have the `resizeToContents` setting
        model = self.model
        self._hasResize = "resizeToContents" in model.copyable_trait_names()
        if self._hasResize:
            resize_action = QAction("Resize To Contents", table_widget)
            resize_action.triggered.connect(self._resize_contents)
            resize_action.setCheckable(True)
            resize_action.setChecked(self.model.resizeToContents)
            table_widget.addAction(resize_action)

        return table_widget

    def set_read_only(self, ro):
        widget = self._table_widget
        if ro:
            self._readonly = True
            widget.setEditTriggers(QAbstractItemView.NoEditTriggers)
            widget.setAlternatingRowColors(True)
            if self.hasCustomMenu:
                widget.setContextMenuPolicy(Qt.CustomContextMenu)
                widget.customContextMenuRequested.connect(self._custom_menu)
                widget.setSelectionMode(QAbstractItemView.SingleSelection)
            else:
                widget.setSelectionMode(QAbstractItemView.NoSelection)
        else:
            self._readonly = False
            widget.setSelectionMode(QAbstractItemView.SingleSelection)
            widget.setSelectionBehavior(QAbstractItemView.SelectRows)
            flags = (QAbstractItemView.DoubleClicked
                     | QAbstractItemView.AnyKeyPressed
                     | QAbstractItemView.SelectedClicked)
            widget.setEditTriggers(flags)
            widget.setContextMenuPolicy(Qt.CustomContextMenu)
            # A reconfigurable table element always has a menu
            handler = (self._custom_menu if self.hasCustomMenu
                       else self._context_menu)
            widget.customContextMenuRequested.connect(handler)
            widget.setAcceptDrops(True)
        widget.set_read_only(self._readonly)
        widget.setFocusPolicy(Qt.NoFocus if ro else Qt.StrongFocus)
        if self._item_model is not None:
            self._item_model.set_readonly(self._readonly)
            # We must set the delegates firstly on readOnly information!
            self.create_delegates()

    def binding_update(self, proxy):
        binding = proxy.binding
        if binding is not None:
            has_schema = not binding.row_schema.empty()
            if has_schema:
                self._set_bindings(binding)
                if self._hasResize:
                    self._set_table_resize_mode(self.model)

    def value_update(self, proxy):
        value = get_editor_value(proxy, [])
        if not value:
            self._item_model.clear_model()
            return

        # Remove or add rows if necessary
        row_count = self._item_model.rowCount()
        if row_count > len(value):
            start = len(value) - 1
            count = row_count - len(value)
            self._item_model.removeRows(start, count, QModelIndex(),
                                        from_device=True)
        elif row_count < len(value):
            start = row_count
            count = len(value) - row_count
            self._item_model.insertRows(start, count, QModelIndex(),
                                        from_device=True)

        self._item_model.updateData(value)

    def state_update(self, proxy):
        """A change in a state update will only disable reconfigurable tables
        """
        if not self._readonly or self._table_buttons:
            enable = is_proxy_allowed(proxy)
            self.widget.setEnabled(enable)

    def destroy_widget(self):
        if self._item_model is not None:
            self._item_model.setParent(None)
            self._item_model = None
        if self.widget:
            self.widget.setParent(None)
            self.widget = None

    def setEnabled(self, enable):
        """A change in access level happens is forward to the widget

        If the table has table buttons, we always enable or disable
        """
        if self._table_buttons:
            self.widget.setEnabled(enable)
        else:
            super().setEnabled(enable)

    # ---------------------------------------------------------------------
    # Subclass Methods

    def create_delegates(self):
        """Subclass method to set the table delegates in the table element"""
        self.setTableDelegates({})

    def custom_menu(self, pos):
        """Subclass method for own custom menu if ``hasCustomMenu`` is `True`

        :param: pos: The position of the context menu event
        """

    def createModel(self, item_model):
        """Subclass the `createModel` to create a filter model

        Note: This method will fade out and stays for backward compatibility
        """
        return self.createFilterModel(item_model)

    def createFilterModel(self, item_model):
        """Subclass the `createFilterModel` to create a filter model

        Note: Future use with 2.15
        """
        return item_model

    # ---------------------------------------------------------------------
    # Public interface

    def setTableDelegates(self, delegates):
        """Set all table delegates on the table controller

        If no delegates are specified for a column, a delegate is derived
        from the column binding. Use this method in `create_delegates`.

        :param delegates: Dictionary {column: delegate}

        Note: This method was added with Karabo 2.16.X
        """
        assert isinstance(delegates, dict)

        bindings = self._bindings
        keys = bindings.keys()
        get_delegate = (get_display_delegate if self._readonly
                        else get_table_delegate)
        self._table_buttons = False
        for column, key in enumerate(keys):
            delegate = delegates.get(column)
            if delegate is None:
                binding = bindings[key]
                delegate = get_delegate(self.proxy, binding,
                                        self._table_widget)
            if isinstance(delegate, TableButtonDelegate):
                self._table_buttons = True
            self._table_widget.setItemDelegateForColumn(column, delegate)

    def getModelData(self, row, column):
        """Get data from the sourceModel with `row` and `column`

        :param row: row of the data
        :param column: column of the data

        Note: Function added with Karabo > 2.13.X

        Method changed in 2.16.X to account for filtering.
        Please use `get_model_data` in the future.
        """
        return self.tableModel().get_model_data(row, column)

    get_model_data = getModelData

    def getInstanceId(self):
        """Retrieve the `instanceId` of the root proxy of this controller

        Note: Function added with Karabo > 2.13.X
        """
        return self.proxy.root_proxy.device_id

    def columnIndex(self, key):
        """Retrieve the column index for a `key` in the schema

        Note: Function added with Karabo 2.16.X
        """
        bindings = list(self._bindings.keys())
        try:
            return bindings.index(key)
        except ValueError:
            return None

    def columnKey(self, column):
        """Retrieve the schema key for a `column` index

        Note: Function added with Karabo 2.16.X
        """
        bindings = list(self._bindings.keys())
        try:
            return bindings[column]
        except IndexError:
            return None

    def currentIndex(self):
        """Convenience method to get the currentIndex of the selection"""
        return self._table_widget.selectionModel().currentIndex()

    def tableModel(self):
        """Return the `model` of the table controller

        Note: Added with Karabo 2.16.X
        """
        return self._table_widget.model()

    def sourceModel(self):
        """Return the `sourceModel` of the table controller"""
        return self._item_model

    def tableWidget(self):
        """Return the table widget of the controller"""
        return self._table_widget

    def getBindings(self):
        """Access for table bindings of the controller"""
        return self._bindings

    def isReadOnly(self):
        """Access for readOnly property of the table controller"""
        return self._readonly

    # ---------------------------------------------------------------------
    # Controller and Actions Slots (Public)

    def add_row(self):
        row = self.currentIndex().row()
        self._item_model.add_row(row)

    def add_row_below(self):
        row = self.currentIndex().row() + 1
        self._item_model.add_row(row)

    def duplicate_row(self):
        row = self.currentIndex().row()
        self._item_model.duplicate_row(row)

    def move_row_up(self):
        row = self.currentIndex().row()
        self._item_model.move_row_up(row)
        self._table_widget.selectRow(row - 1)

    def move_row_down(self):
        row = self.currentIndex().row()
        self._item_model.move_row_down(row)
        self._table_widget.selectRow(row + 1)

    def remove_row(self):
        index = self.currentIndex()
        self._item_model.removeRows(index.row(), 1, QModelIndex())

    def get_basic_menu(self):
        """Used by subclassed controller to get the basic menu

        Note: Added in Karabo 2.16.X
        """
        menu = QMenu(parent=self._table_widget)
        if self.isReadOnly():
            return menu

        index = self.currentIndex()
        if index.isValid():
            key = list(self._bindings.keys())[index.column()]
            binding = self._bindings[key]
            if is_writable_binding(binding):
                set_default_action = menu.addAction("Set Cell Default")
                set_default_action.triggered.connect(self._set_index_default)
                menu.addSeparator()

            up_action = menu.addAction(icons.arrowFancyUp, "Move Row Up")
            up_action.triggered.connect(self.move_row_up)
            down_action = menu.addAction(icons.arrowFancyDown, "Move Row Down")
            down_action.triggered.connect(self.move_row_down)
            menu.addSeparator()

            add_action = menu.addAction(icons.add, "Add Row below")
            add_action.triggered.connect(self.add_row_below)
            du_action = menu.addAction(icons.editCopy, "Duplicate Row below")
            du_action.triggered.connect(self.duplicate_row)

            remove_action = menu.addAction(icons.delete, "Delete Row")
            remove_action.triggered.connect(self.remove_row)

            # Check for min and max size of the table
            row_count = self._item_model.rowCount()
            attributes = self.proxy.binding.attributes
            min_size = attributes.get(KARABO_SCHEMA_MIN_SIZE)
            max_size = attributes.get(KARABO_SCHEMA_MAX_SIZE)
            add_row = True if max_size is None else row_count + 1 <= max_size
            rm_row = True if min_size is None else row_count - 1 >= min_size

            # Set actions enabled or disabled!
            num_row = self.tableWidget().model().rowCount() - 1
            up_action.setEnabled(index.row() > 0)
            down_action.setEnabled(index.row() < num_row)
            add_action.setEnabled(add_row)
            du_action.setEnabled(add_row)
            remove_action.setEnabled(rm_row)

            if (is_writable_binding(binding, StringBinding)
                    and not has_options(binding)):
                menu.addSeparator()
                device_action = menu.addAction(
                    icons.deviceInstance, "Set Topology DeviceId")
                device_action.triggered.connect(self._device_action)

        else:
            add_action = menu.addAction(icons.add, "Add Row below")
            add_action.triggered.connect(self.add_row_below)

        return menu

    # ---------------------------------------------------------------------
    # Action Slots - Private

    def _device_action(self):
        model = self.tableWidget().model()
        index = model.index_ref(self.currentIndex())
        dialog = TopologyDeviceDialog(parent=self.widget)
        if dialog.exec() == QDialog.Accepted:
            device = dialog.device_id
            self.sourceModel().setData(index, device)

    def _set_index_default(self):
        model = self.tableWidget().model()
        index = model.index_ref(self.currentIndex())
        key = list(self._bindings.keys())[index.column()]
        binding = self._bindings[key]
        default_value = get_default_value(binding, force=True)
        self._item_model.setData(index, default_value, role=Qt.EditRole)

    def _resize_contents(self):
        self.model.resizeToContents = not self.model.resizeToContents
        self._set_table_resize_mode(self.model)

    def _set_table_resize_mode(self, model):
        mode = (QHeaderView.ResizeToContents if model.resizeToContents
                else QHeaderView.Interactive)

        length = len(self.getBindings().keys())
        for index in range(length - 1):
            self._table_widget.horizontalHeader().setSectionResizeMode(
                index, mode)
        self._table_widget.horizontalHeader().setStretchLastSection(True)

    def _context_menu(self, pos):
        """The custom context menu of a reconfigurable table element"""
        selection_model = self._table_widget.selectionModel()
        if selection_model is None:
            # Note: We did not yet receive a schema and thus have no table and
            # selection model!
            return

        menu = self.get_basic_menu()
        menu.exec(self._table_widget.viewport().mapToGlobal(pos))

    # Private interface
    # ---------------------------------------------------------------------

    def _on_user_edit(self, data):
        """Callback method used by `self._item_model` when data changes"""
        if self.proxy.binding is None:
            return
        self.proxy.edit_value = data

    def _set_bindings(self, binding):
        """Configure the column schema hashes and keys

        The schema must not be `None` and is protected when calling this func.
        """
        if self._item_model is not None:
            self._item_model.setParent(None)
            self._item_model = None

        self._bindings = binding.bindings

        source_model = self.tableModelClass(binding, self._on_user_edit,
                                            self._table_widget)
        self._item_model = source_model
        self._item_model.set_readonly(self._readonly)
        model = self.createModel(self._item_model)
        self._table_widget.setModel(model)
        self.create_delegates()

    def _custom_menu(self, pos):
        """Private method to direct to the custom context menu"""
        selection_model = self._table_widget.selectionModel()
        if selection_model is None:
            return

        return self.custom_menu(pos)


# --------------------------------------------------------------------------
# Filter Table Controller


class BaseFilterTableController(BaseTableController):
    searchLabel = Instance(QLineEdit)
    columnCombo = Instance(QComboBox)

    def create_widget(self, parent):
        table_widget = super().create_widget(parent)
        widget = QWidget(parent)
        widget.addActions(table_widget.actions())

        widget_layout = QVBoxLayout()
        widget_layout.setContentsMargins(0, 0, 0, 0)
        widget_layout.setSizeConstraint(QLayout.SetNoConstraint)

        hor_layout = QHBoxLayout()
        hor_layout.setContentsMargins(0, 0, 0, 0)
        hor_layout.setSizeConstraint(QLayout.SetNoConstraint)

        self.columnCombo = QComboBox(widget)
        self.columnCombo.setVisible(False)
        self.columnCombo.setToolTip("Active search column")

        self.searchLabel = QLineEdit(widget)
        self.searchLabel.setToolTip(
            f"Persisted search column: {self.model.filterKeyColumn}")

        clear_button = QPushButton("Clear Filter", parent=widget)
        clear_button.clicked.connect(self.searchLabel.clear)

        hor_layout.addWidget(self.columnCombo)
        hor_layout.addWidget(self.searchLabel)
        hor_layout.addWidget(clear_button)

        # Complete widget layout and return widget
        widget_layout.addLayout(hor_layout)
        widget_layout.addWidget(table_widget)
        widget.setLayout(widget_layout)

        ac_column_filter = QAction("Set Default Filter Column", table_widget)
        ac_column_filter.triggered.connect(self._change_filter_column)
        widget.addAction(ac_column_filter)

        # Widgets in extensions might not have this model setting
        traits_names = self.model.copyable_trait_names()
        if "sortingEnabled" in traits_names:
            enabled = self.model.sortingEnabled
            if enabled:
                table_widget.setSortingEnabled(True)
                table_widget.sortByColumn(0, Qt.AscendingOrder)

            ac_sort = QAction("Sorting Enabled", table_widget)
            ac_sort.setCheckable(True)
            ac_sort.setChecked(enabled)
            ac_sort.triggered.connect(self._change_sorting_enabled)
            widget.addAction(ac_sort)

        if "showFilterKeyColumn" in traits_names:
            enabled = self.model.showFilterKeyColumn
            self.columnCombo.setVisible(enabled)
            ac_show = QAction("Show Filter Column Toggle", table_widget)
            ac_show.setCheckable(True)
            ac_show.setChecked(enabled)
            ac_show.triggered.connect(self._change_show_combo)
            widget.addAction(ac_show)

        return widget

    def createFilterModel(self, item_model):
        """Create the filter model for the table"""
        filter_model = TableSortFilterModel(self.tableWidget())
        filter_model.setSourceModel(item_model)
        key = self.model.filterKeyColumn
        filter_model.setFilterKeyColumn(key)
        filter_model.setFilterRole(Qt.DisplayRole)
        filter_model.setFilterCaseSensitivity(False)
        filter_model.setFilterFixedString("")
        self.searchLabel.textChanged.connect(filter_model.setFilterFixedString)
        self.columnCombo.currentIndexChanged.connect(
            filter_model.setFilterKeyColumn)

        return filter_model

    def binding_update(self, proxy):
        super().binding_update(proxy)
        model = self.tableModel()
        columns = [model.headerData(col, Qt.Horizontal, Qt.DisplayRole)
                   for col in range(model.columnCount())]
        with SignalBlocker(self.columnCombo):
            self.columnCombo.clear()
            self.columnCombo.addItems(columns)
            self.columnCombo.setCurrentIndex(self.model.filterKeyColumn)

    def _change_filter_column(self):
        max_col = len(self.getBindings())
        column, ok = QInputDialog.getInt(
            self.widget, "Filter Column", "Filter Key Column:",
            value=self.model.filterKeyColumn,
            min=0, max=max_col)
        if ok:
            self.model.filterKeyColumn = column
            filter_model = self.tableWidget().model()
            filter_model.setFilterKeyColumn(column)
            self.searchLabel.setToolTip(f"Persisted search column: {column}")
            with SignalBlocker(self.columnCombo):
                self.columnCombo.setCurrentIndex(column)

    def _change_sorting_enabled(self):
        enabled = not self.model.sortingEnabled
        self.model.sortingEnabled = enabled
        self.tableWidget().setSortingEnabled(enabled)
        if enabled:
            self.tableWidget().sortByColumn(0, Qt.AscendingOrder)

    def _change_show_combo(self):
        enabled = not self.model.showFilterKeyColumn
        self.model.showFilterKeyColumn = enabled
        self.columnCombo.setVisible(enabled)

    # ------------------------------------------------------------------------
    # Action interface

    def sourceRow(self):
        """Method to retrieve the selected row of the source model"""
        index = self.currentIndex()
        return self.tableModel().mapToSource(index).row()

    def add_row(self):
        row = self.sourceRow()
        self._item_model.add_row(row)

    def add_row_below(self):
        row = self.sourceRow()
        self._item_model.add_row(row)

    def duplicate_row(self):
        row = self.sourceRow()
        self._item_model.duplicate_row(row)

    def move_row_up(self):
        row = self.sourceRow()
        select = self.currentIndex().row() - 1
        self._item_model.move_row_up(row)
        self._table_widget.selectRow(select)

    def move_row_down(self):
        row = self.sourceRow()
        select = self.currentIndex().row() + 1
        self._item_model.move_row_down(row)
        self._table_widget.selectRow(select)

    def remove_row(self):
        row = self.sourceRow()
        self._item_model.removeRows(row, 1, QModelIndex())



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from traits.api import Undefined

from karabo.common import const
from karabo.native import Hash, NodeType

from . import binding_types as types
from .util import attr_fast_deepcopy


def build_binding(schema, existing=None):
    """Given a schema object, build an object binding which matches the
    structure defined in the schema. All leaf nodes are built with an initial
    value of Undefined.
    """
    # Use `existing` or create a new instance.
    binding = existing or types.BindingRoot()
    binding.class_id = schema.name
    root_namespace = binding.value
    # Clear the namespace
    root_namespace.clear_namespace()
    # Fill it back in recursively
    for key, value, attrs in schema.hash.iterall():
        node = _build_node(value, attrs)
        setattr(root_namespace, key, node)

    # Let any listeners know
    binding.schema_update = True

    return binding


def _build_node(value, attrs):
    """Recursively build a single node (and its children) in an object binding
    """
    # Always deal with a copy of the attributes
    # (to allow bindings to modify)
    attrs = attr_fast_deepcopy(attrs)

    def _build_subnamespace(hashval, item_type):
        assert isinstance(hashval, Hash)
        namespace = types.BindingNamespace(item_type=types.BaseBinding)
        for subname, subvalue, subattrs in hashval.iterall():
            subnode = _build_node(subvalue, subattrs)
            setattr(namespace, subname, subnode)
        return namespace

    node_type = attrs[const.KARABO_SCHEMA_NODE_TYPE]
    if node_type == NodeType.Leaf:
        value_type = attrs[const.KARABO_SCHEMA_VALUE_TYPE]
        binding_factory = _BINDING_MAP[value_type]
        return binding_factory(attributes=attrs, value=Undefined)

    elif node_type == NodeType.Node:
        displayType = attrs.get(
            const.KARABO_SCHEMA_DISPLAY_TYPE, '').split('|')[0]
        namespace = _build_subnamespace(value, types.BaseBinding)
        if displayType in _NODE_BINDING_MAP:
            # Certain display types get a different binding class
            binding_factory = _NODE_BINDING_MAP[displayType]
            return binding_factory(value=namespace, attributes=attrs)
        return types.NodeBinding(value=namespace, attributes=attrs)

    raise RuntimeError(f"Not supported value type {node_type}")


# ----------------------------------------------------------------------------
# A dict to help with valueType -> binding type mapping
_BINDING_MAP = {
    'BOOL': types.BoolBinding,
    'CHAR': types.CharBinding,
    'INT8': types.Int8Binding,
    'UINT8': types.Uint8Binding,
    'INT16': types.Int16Binding,
    'UINT16': types.Uint16Binding,
    'INT32': types.Int32Binding,
    'UINT32': types.Uint32Binding,
    'INT64': types.Int64Binding,
    'UINT64': types.Uint64Binding,
    'FLOAT': types.FloatBinding,
    'DOUBLE': types.FloatBinding,
    'STRING': types.StringBinding,
    'HASH': types.HashBinding,
    'SCHEMA': types.SchemaBinding,
    'NONE': types.NoneBinding,
    'BYTE_ARRAY': types.ByteArrayBinding,
    'VECTOR_BOOL': types.VectorBoolBinding,
    'VECTOR_CHAR': types.VectorCharBinding,
    'VECTOR_INT8': types.VectorInt8Binding,
    'VECTOR_UINT8': types.VectorUint8Binding,
    'VECTOR_INT16': types.VectorInt16Binding,
    'VECTOR_UINT16': types.VectorUint16Binding,
    'VECTOR_INT32': types.VectorInt32Binding,
    'VECTOR_UINT32': types.VectorUint32Binding,
    'VECTOR_INT64': types.VectorInt64Binding,
    'VECTOR_UINT64': types.VectorUint64Binding,
    'VECTOR_DOUBLE': types.VectorDoubleBinding,
    'VECTOR_FLOAT': types.VectorFloatBinding,
    'VECTOR_STRING': types.VectorStringBinding,
    'VECTOR_HASH': types.VectorHashBinding,
    'VECTOR_NONE': types.VectorNoneBinding,
}
_NODE_BINDING_MAP = {
    'ImageData': types.ImageBinding,
    'WidgetNode': types.WidgetNodeBinding,
    'Image': types.ImageBinding,
    'NDArray': types.NDArrayBinding,
    'Slot': types.SlotBinding,
    'OutputChannel': types.PipelineOutputBinding,
    'Table': types.TableBinding,
}



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on November 4, 2011
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtCore import Slot
from qtpy.QtGui import QCursor
from qtpy.QtWidgets import QTabBar, QTabWidget

from .placeholderpanel import PlaceholderPanel
from .utils import get_panel_icon


class PanelContainer(QTabWidget):
    """A container for ``BasePanelWidget`` instances
    """
    def __init__(self, title, parent, handle_empty=False):
        super().__init__(parent)
        self.setWindowTitle(title)
        self.panel_set = set()
        self.handle_empty = handle_empty
        self.minimized = False
        self.maximized = False
        self.tabCloseRequested.connect(self.onCloseTab)

        self._add_placeholder()
        self._update_tabs_closable()

    # ----------------------------------------------------------------------
    # public methods

    def addPanel(self, panel):
        """This function gets a panel and adds it to this container.
        """
        self._remove_placeholder()

        # XXX: Circular references hurrah!
        panel.attach_to_container(self)

        # if the container is currently maximized, quitely add the new
        # tab in the background
        if not self.maximized:
            index = self.addTab(panel, panel.windowTitle())
            panel.index = index

        # The container might have been hidden
        # XXX: currently no way to know if other container is maximized,
        # so that we can decide if this container should stay hidden.
        if self.count() == 1:
            self.show()

        # Store panel in set to keep it alive for un/dock event!!!
        self.panel_set.add(panel)
        self._update_tabs_closable()

    def removePanel(self, panel):
        # trigger this to class closeEvent for panels and destroy connections
        # properly
        panel.close()

        # Check if panel is undocked. Remove it with the corresponding function
        index = self.indexOf(panel)
        if index < 0:
            self.removeUndockedPanel(panel)
            return

        # Then go about the removal of the tab
        self.removeTab(index)
        panel.setParent(None)
        self.panel_set.remove(panel)
        if self.maximized:
            # restore other tabs before remove it
            panel.onMinimize()
        self._update_tabs_closable()

        self._add_placeholder()
        if self.count() == 0:
            self.hide()

    def removeUndockedPanel(self, panel):
        """Remove an undocked panel from tracking

        This method is called by the panel itself as only closing signals
        are emitted for docked panels
        """
        # Check if the panel is in the process or has been removed already.
        if panel not in self.panel_set:
            return

        index = self.indexOf(panel)
        assert index < 0 and not panel.is_docked

        self.panel_set.remove(panel)
        panel.setParent(None)
        panel.deleteLater()

    def minimize(self, minimized):
        """Minimize/unminimize a tab container.
        """
        self.minimized = minimized
        if minimized:
            self.hide()
        elif self.count() > 0:
            # Only show if the container is NOT empty!
            self.show()

    def dock(self, panel):
        if panel.parent() is None:
            self._remove_placeholder()

            index = self.insertTab(panel.index, panel, panel.windowTitle())
            panel.is_docked = True

            for i in range(self.count()):
                if self.widget(i) is not None:
                    self.widget(i).index = i

            if self.maximized:
                # newly docked tab shouldn't appear and steal the focus if
                # the container is maximized with other panels
                self.removeTab(index)
            else:
                self.setCurrentIndex(index)
            self._update_tabs_closable()

            if not self.minimized:
                self.show()

    def undock(self, panel):
        if panel.parent() is not None:
            # NOTE: If panel is maximized, parents for links appear to be None!
            panel.setParent(None)
        panel.is_docked = False
        panel.move(QCursor.pos())
        panel.show()

        self._add_placeholder()
        if self.count() == 0:
            self.hide()

    def insert_panels_after_maximize(self, current_index):
        """When maximized, all panels are removed except for the selected one.
        This undoes that when the panel/container is minimized.
        """
        self.removeTab(0)
        # Add the tabs back to the container in sorted order
        panels = sorted(self.panel_set, key=lambda x: x.index)
        for pan in panels:
            if not pan.is_docked:
                continue
            self.insertTab(pan.index, pan, pan.windowTitle())

        self.setCurrentIndex(current_index)
        self._hide_close_buttons()

    def update_tab_text_color(self, panel, color):
        if panel.is_docked:
            index = self.indexOf(panel)
            self._set_tab_text_color(index, panel, color)

    def switch_to_panel(self, panel):
        if panel.parent() is not None and not self.maximized:
            # The panel is docked and is added as a tab. We switch to it.
            self.setCurrentIndex(panel.index)
        else:
            # The panel is undocked. We put the window in front.
            panel.raise_()

    # --------------------------------------
    # Qt Overrides

    def addTab(self, widget, label):
        index = super().addTab(widget, label)
        self._set_tab_text_color(index, widget, widget.tab_text_color)
        self._set_tab_icon(index, widget)
        return index

    def insertTab(self, index, widget, label):
        index = super().insertTab(index, widget, label)
        self._set_tab_text_color(index, widget, widget.tab_text_color)
        self._set_tab_icon(index, widget)
        return index

    # ----------------------------------------------------------------------
    # slots

    @Slot(int)
    def onCloseTab(self, index):
        # Get panel, which is about to be closed
        panel = self.widget(index)
        # Close panel (if possible) before removing it from tab
        if not panel.close():
            return
        # Remove panel from tab
        self.removeTab(index)
        panel.setParent(None)
        self.panel_set.remove(panel)

        if self.maximized:
            panel.onMinimize()

        self._add_placeholder()
        # if all tabs are closed, hide myself
        if self.count() == 0:
            self.hide()

        panel.deleteLater()

    # ----------------------------------------------------------------------
    # private methods

    def _add_placeholder(self):
        """Add a placeholder if needed
        """
        if self.handle_empty and self.count() == 0:
            placeholder = PlaceholderPanel()
            self.addTab(placeholder, placeholder.windowTitle())
            self._update_tabs_closable()

    def _remove_placeholder(self):
        """Remove the placeholder if needed
        """
        if self.handle_empty and self.count() == 1:
            if isinstance(self.widget(0), PlaceholderPanel):
                self.removeTab(0)

    def _set_tab_text_color(self, index, widget, color=None):
        if color is not None and color.isValid():
            tab_bar = self.tabBar()
            tab_bar.setTabTextColor(index, color)

    def _set_tab_icon(self, index, widget):
        icon = get_panel_icon(widget)
        if icon is not None:
            self.setTabIcon(index, icon)

    def _update_tabs_closable(self):
        if self.count() > 1:
            self.setTabsClosable(True)
        elif self.count() == 1:
            is_placeholder = isinstance(self.widget(0), PlaceholderPanel)
            self.setTabsClosable(not is_placeholder)
        self._hide_close_buttons()

    def _hide_close_buttons(self):
        """Remove the close button non-closable panels
        """
        # use index instead of loop through set, because need the index
        # for tabBar button anyway.
        i = self.count()
        while i > -1:
            i -= 1
            widget = self.widget(i)
            if widget is not None and not widget.allow_closing:
                # some system put close button on the left
                for p in (QTabBar.LeftSide, QTabBar.RightSide):
                    bt = self.tabBar().tabButton(i, p)
                    if bt is not None:
                        bt.resize(0, 0)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from karabo.native import Bool, Configurable
from karabogui.testing import (
    check_renderer_against_svg, get_class_property_proxy, set_proxy_value)

from ..checkbox import CHECKED, UNCHECKED, DisplayCheckBox


class Object(Configurable):
    prop = Bool(defaultValue=True)


def test_checkbox_set_value(gui_app):
    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, 'prop')
    controller = DisplayCheckBox(proxy=proxy)
    controller.create(None)

    set_proxy_value(proxy, 'prop', True)
    check_renderer_against_svg(controller.widget.renderer(),
                               CHECKED)

    set_proxy_value(proxy, 'prop', False)
    check_renderer_against_svg(controller.widget.renderer(),
                               UNCHECKED)

    controller.destroy()
    assert controller.widget is None



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from karabo.native import Configurable, Int32, Node
from karabogui.testing import get_class_property_proxy

from ..widgetnode import DisplayWidgetNode


class DataNode(Configurable):
    displayType = "WidgetNode|DataNode"
    value = Int32(defaultValue=32)


class Object(Configurable):
    node = Node(DataNode)


def test_widgetnode_version(gui_app):
    # setup
    schema = Object.getClassSchema()
    widget_node = get_class_property_proxy(schema, 'node')
    controller = DisplayWidgetNode(proxy=widget_node)
    controller.create(None)

    # test body
    assert controller.widget.text() == "WidgetNode\nNodeType: DataNode"

    # teardown
    controller.destroy()
    assert controller.widget is None



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from karabo.common.project.api import ProjectModel
from karabogui.panels.projectpanel import ProjectPanel
from karabogui.singletons.project_model import ProjectViewItemModel
from karabogui.testing import singletons


def test_project_panel(gui_app, mocker):
    qt_model = ProjectViewItemModel()
    mediator = mocker.Mock()
    with singletons(project_model=qt_model, mediator=mediator):
        panel = ProjectPanel()
        assert qt_model.root_model is None
        assert repr(panel) == "<ProjectPanel project=None>"
        project_model = ProjectModel(simple_name="Test")
        qt_model.root_model = project_model
        assert repr(panel) == "<ProjectPanel project=Test>"



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on April 28, 2020
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy import uic
from qtpy.QtCore import QSize, Qt, Slot
from qtpy.QtGui import QColor, QFont, QIcon, QPixmap, QTextFormat
from qtpy.QtWidgets import QApplication, QColorDialog, QDialog, QTextEdit

from karabogui.dialogs.font_dialog import FontDialog
from karabogui.fonts import get_alias_from_font, get_qfont, substitute_font

from .utils import get_dialog_ui

HIGHLIGHT_COLOR = QColor(Qt.yellow).lighter(180)
GREY = "#d3d3d3"
BUTTON_FONT_SIZE = 10


class StickerDialog(QDialog):

    def __init__(self, model=None, parent=None):
        super().__init__(parent)
        uic.loadUi(get_dialog_ui('stickerdialog.ui'), self)

        self.model = model.clone_traits()
        substitute_font(self.model)

        # Set default font if the model font is empty
        if self.model.font:
            self.text_font = get_qfont(self.model.font, adjust_size=False)
        else:
            self.text_font = QApplication.font()
            self.model.font = self.text_font.toString()

        self.set_text_font_button()
        self.set_text_color_button()
        self.set_text_background_button()
        self.cbBackground.setChecked(self.model.background != 'transparent')

        self.update_plain_widget()

    def update_plain_widget(self):
        model = self.model
        self.leText.setFixedSize(QSize(int(model.width), int(model.height)))
        self.leText.setPlainText(model.text)
        sheet = []
        font_string = get_qfont(model.font, adjust_size=True).toString()
        sheet.append(f'qproperty-font: "{font_string}";')
        sheet.append(f'color: "{model.foreground}";')
        sheet.append(f'background-color: "{model.background}";')

        # Add borders
        bg = model.background
        if bg == 'transparent':
            bg = GREY
        color = QColor(bg).darker(120).name()
        sheet.append('border: 1px solid;')
        sheet.append('border-top: 5px solid;')
        sheet.append(f'border-color: {color};')

        self.leText.setStyleSheet("QPlainTextEdit {{ {} }}".format("".join(
            sheet)))

    def set_text_font_button(self):
        """
        Update the  text and font family of the button. The font size
        should not be changed in order to keep the button size fixed.
        """
        font = QFont(self.text_font)
        font.setPointSize(BUTTON_FONT_SIZE)
        self.pbFont.setFont(font)
        family = get_alias_from_font(self.text_font.family())
        button_text = f"{family}, {self.text_font.pointSize()}pt"
        self.pbFont.setText(button_text)

    def set_text_color_button(self):
        pixmap = QPixmap(24, 16)
        pixmap.fill(QColor(self.model.foreground))
        self.pbTextColor.setIcon(QIcon(pixmap))

    def set_text_background_button(self):
        pixmap = QPixmap(24, 16)
        pixmap.fill(QColor(self.model.background))
        self.pbBackground.setIcon(QIcon(pixmap))

    # ----------------------------------------------------------------------
    # Qt Slots

    @Slot()
    def on_leText_cursorPositionChanged(self):
        extra_selections = []
        selection = QTextEdit.ExtraSelection()
        selection.format.setBackground(HIGHLIGHT_COLOR)
        selection.format.setProperty(QTextFormat.FullWidthSelection, True)
        selection.cursor = self.leText.textCursor()
        selection.cursor.clearSelection()
        extra_selections.append(selection)
        self.leText.setExtraSelections(extra_selections)

    @Slot()
    def on_leText_textChanged(self):
        self.model.text = self.leText.toPlainText()

    @Slot()
    def on_pbFont_clicked(self):
        dialog = FontDialog(self.text_font, parent=self)
        if dialog.exec() == QDialog.Accepted:
            self.text_font = dialog.qfont
            self.model.font = self.text_font.toString()
            self.set_text_font_button()
            self.update_plain_widget()

    @Slot()
    def on_pbTextColor_clicked(self):
        color = QColorDialog.getColor(QColor(self.model.foreground))
        if color.isValid():
            self.model.foreground = color.name()
            self.set_text_color_button()
            self.update_plain_widget()

    @Slot(bool)
    def on_cbBackground_toggled(self, checked):
        self.pbBackground.setEnabled(checked)
        if not checked:
            self.model.background = 'transparent'
            self.set_text_background_button()
        self.update_plain_widget()

    @Slot()
    def on_pbBackground_clicked(self):
        color = QColorDialog.getColor(QColor(self.model.background))
        if color.isValid():
            self.model.background = color.name()
            self.set_text_background_button()
        self.update_plain_widget()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from unittest import mock

from qtpy.QtCore import QSize

from karabogui.testing import GuiTestCase
from karabogui.widgets.toolbar import ToolBar


class TestPopUp(GuiTestCase):

    def test_default(self):
        widget = ToolBar()
        assert widget.windowTitle() == ""
        assert widget.iconSize() == QSize(19, 19)
        assert widget.isMovable() is False
        assert widget.isFloatable() is True

        # title test
        widget = ToolBar("new toolbar")
        assert widget.windowTitle() == "new toolbar"

        with mock.patch.object(widget, 'addWidget') as mock_method:
            # add_expander will add a widget to the toolbar. This way
            # we can test if that method was called.
            mock_method.assert_not_called()
            widget.add_expander()
            mock_method.assert_called_once()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np

from karabo.native import Configurable, Hash, NDArray, UInt32, VectorDouble
from karabogui.testing import get_class_property_proxy, set_proxy_hash

from ..arrays import get_array_data


class Object(Configurable):
    prop = NDArray(
        displayedName="One dimensional array",
        dtype=UInt32,
        shape=(10,))


class TwoObject(Configurable):
    prop = NDArray(
        displayedName="Two dimensional array",
        dtype=UInt32,
        shape=(10, 2))


def test_array():
    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, 'prop')
    value = np.array(list(range(10)), dtype="uint32")
    array_hash = Hash('type', 14,
                      'data', value.tobytes())
    h = Hash('prop', array_hash)
    set_proxy_hash(proxy, h)
    property_value, _ = get_array_data(proxy)
    np.testing.assert_almost_equal(property_value, value)

    schema = TwoObject.getClassSchema()
    proxy = get_class_property_proxy(schema, 'prop')
    value = np.array([[np.random.randint(0, 50) for _ in range(10)]
                      for _ in range(2)], dtype="uint32")
    array_hash = Hash('type', 14,
                      'data', value.tobytes())
    h = Hash('prop', array_hash)
    set_proxy_hash(proxy, h)
    property_value, _ = get_array_data(proxy)
    # XXX: Traits does casting to a different 1 dimensional shape!
    assert property_value is not None


def test_vector():
    """Test the `get_array` function with vectors"""

    class VectorObject(Configurable):
        prop = VectorDouble(
            displayedName="Vector Double")

    schema = VectorObject.getClassSchema()
    proxy = get_class_property_proxy(schema, 'prop')
    property_value, _ = get_array_data(proxy)
    assert property_value is None
    property_value, _ = get_array_data(proxy, default=[])
    assert property_value == []
    value = [1, 2, 3, 4]
    h = Hash('prop', value)
    set_proxy_hash(proxy, h)
    property_value, _ = get_array_data(proxy)
    np.testing.assert_almost_equal(property_value, value)
